#!/usr/bin/env python
# coding: utf-8

'''
A proprietary module for facilitating the use of selenium
Авторский модуль для упрощения некоторых оперций в selenium
'''
# import sys
# sys.path.append(r"C:\Users\Alexey\Dropbox\Мои\RAnDan\myModules")

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from selenium import webdriver
        from selenium.webdriver.common.by import By # для поиска элементов HTML-кода
        from selenium.webdriver.support import expected_conditions
        from selenium.webdriver.support.ui import WebDriverWait
        import re, selenium.common.exceptions, time
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break
            
def blockSearch(attemptsMax, driver, text, xPathS):
    block = None 
    trCounter = 1
    while True:
        # print('xPath:', xPathS[0] + str(trCounter) + xPathS[1]) # для отладки
        try:
            block = driver.find_element(By.XPATH, xPathS[0] + str(trCounter) + xPathS[1]).text
            if text in block: break
            trCounter += 1
        except:
            # print('trCounter:', trCounter) # для отладки
            trCounter += 1
            if trCounter > attemptsMax: break # против бесконечного цикла при пустом блоке страницы
    return block if text in block else None
    
def pathRelative(driver, pathAnchor, pathTarget, pause, textAnchor, textTarget):
    if pathAnchor == None: pathAnchor = pathTarget
    if textAnchor == None: textAnchor = textTarget

    # print('pathAnchor:', pathAnchor) # для отладки
    # print('pathTarget:', pathTarget) # для отладки
    # print('textAnchor:', textAnchor) # для отладки
    # print('textTarget:', textTarget) # для отладки

    if len(re.findall(textAnchor, driver.find_element("tag name", "body").text, re.IGNORECASE)) == 1: # проверить уникальность
        elementAnchor = WebDriverWait(driver, pause).until(expected_conditions.presence_of_element_located((By.XPATH, pathAnchor))) # найти якорь
        # print('elementAnchor:', elementAnchor) # для отладки
        elementTarget = elementAnchor if pathAnchor == pathTarget else elementAnchor.find_element(By.XPATH, pathTarget) # от якоря к кнопке
        # print('elementTarget:', elementTarget) # для отладки
        return elementTarget

def tryerSleeper(attemptsMax, boundarieS, driver, pause, xPathS):
    goS = True
    goC = True
    
    attempt = 1
    while (attempt < attemptsMax) & goC:
        if boundarieS != None:
            for i in range(boundarieS[0], boundarieS[1]): # цикл на случай вариативности дивов
                xPath = xPathS[0] + str(i) + xPathS[1]
                try:
                    driver.find_element(By.XPATH, xPath)
                    goC = False
                    break
                except selenium.common.exceptions.NoSuchElementException:
                    errorDescription = sys.exc_info()
                    print(f'Попытка tryerSleeper № {attempt} . Ошибка:', errorDescription, '          ', end='\r') # , end='\r'
                    time.sleep(pause)
                    attempt += 1
                    if attempt == attemptsMax: goS = False
        else:
            xPath = xPathS[0]
            try:
                driver.find_element(By.XPATH, xPath)
                goC = False
                break
            except selenium.common.exceptions.NoSuchElementException:
                errorDescription = sys.exc_info()
                print(f'Попытка tryerSleeper № {attempt} . Ошибка:', errorDescription, '          ', end='\r') # , end='\r')
                time.sleep(pause)
                attempt += 1
                if attempt == attemptsMax: goS = False
    print('                                                                                                                                                                                                                                                ', end='\r')
    return goS, xPath



#!/usr/bin/env python
# coding: utf-8

'''
A module for editing one dataframe (df_editing) within its specific column based on the same column from another dataframe (df_standard)
'''
# import sys
# sys.path.append(r"C:\Users\Alexey\Dropbox\Мои\RAnDan\myModules")

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from pandas import DataFrame
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

def dictionariesHarmonizer(df_editing, df_standard, columnName):
    df_editing = df_editing.copy() # df_editing -- датафрейм, редактируемый в столбце columnName на основе того же столбца датафрейма df_standard

    # Шаг № 1. Грубая сверка
    df_editing_matching = df_editing[df_editing[columnName].isin(df_standard[columnName])]
    df_editing_new_1 = df_editing[df_editing[columnName].isin(df_standard[columnName]) != True]

    # Шаг № 2. Тонкая сверка
    rowS_detected = [] # только эти строки датафрейма останутся в df_editing_new_1
    df_editing_new_2 = df_editing_new_1.copy()
    elementS_editing = df_editing_new_1[columnName]
    for element_editing in elementS_editing:
        # print('element_editing:', element_editing) # для отладки , end='\r'
        for element_standard in df_standard[columnName]:
            # print('element_standard:', element_standard) # для отладки , end='\r'
            if element_standard in element_editing:
                # print('element_standard in element_editing:', element_standard in element_editing) # для отладки , end='\r'
                rowS_detected.extend(df_editing_new_1[df_editing_new_1[columnName] == element_editing].index)
                df_editing_new_1.loc[df_editing_new_1[columnName] == element_editing, columnName] = element_standard # заменить element_editing на element_standard ,
                    # что обеспечивает совместимость обрабатываемых тут строчек df_editing_new_1 и df_standard
                df_editing_new_2 = df_editing_new_2[df_editing_new_2[columnName] != element_editing]

    rowS_detected = list(set(rowS_detected))
    rowS_detected.sort
    df_editing_new_1 = df_editing_new_1.loc[rowS_detected, :]

    return df_editing_matching, df_editing_new_1, df_editing_new_2
    # df_editing_new_1 -- часть редактируемого датафрейма (df_editing), которая не прошла грубую сверку, но прошла тонкую сверку
    # df_editing_new_2 -- часть редактируемого датафрейма (df_editing), которая не прошла ни грубую, ни тонкую сверку



# country = 'Благовещенск'
# country = 'Египет'
country = 'Индия'
# country = 'Иордания'
# country = 'Кипр'
# country = 'ОАЭ'
# country = 'Саудовская Аравия'
# country = 'Тайланд'
# country = 'Шри-Ланка'
priceTarget = 30000
transferscount = 1
withbaggage = True

from selenium import webdriver
from selenium.webdriver.common.by import By # для поиска элементов HTML-кода
import datetime, pandas, os, sys, time, warnings
import selenium.common.exceptions
pause = 1

def билеты_есть(city, date, datePriceS, driver, pause, priceTarget, url):
    goF = True # взаимодействует с билетов_нет()
    price = 0 # заранее -- на случай отсутствия билетов       
    print('    Предварительно билеты есть                   ') #, end='\r'

    # Отсортировать по увеличению цены

    firstMoment = datetime.datetime.now()
    while goF:
        try:
            driver.find_element(By.XPATH, "//button[@data-qa-file='SortDropdown']")
            sortDropdown = driver.find_element(By.XPATH, "//button[@data-qa-file='SortDropdown']")
            print('    Кнопка сортировки появилась               ') #, end='\r'
            break
        except:
            # print('\n    ', sys.exc_info()[1]) # для отладки
            print('    Жду кнопку сортировки                   ', end='\r') #, end='\r'
            time.sleep(pause)
            lastMoment = datetime.datetime.now()
            diffetenceSeconds = lastMoment - firstMoment
            print('    Ожидание составляет:', diffetenceSeconds.seconds, 'сек.', end='\r') #, end='\r'
            if 'Билетов нет' in driver.find_element("tag name", "body").text:
                driver, goF = билетов_нет(driver, goF, pause, url)
                # print('goF:', goF) # для отладки

    if goF:
        if 'Дешевле' not in sortDropdown.text:
            print('    Меняю сортировку                   ') #, end='\r'
            sortDropdown.click()
            driver.find_element(By.XPATH, "//div[@data-qa-file='item' and contains(., 'Дешевле')]").click()
            time.sleep(pause)
    
        time.sleep(pause * 25) # чтобы успеть посмотреть на цены на странице
    
        # Выгрузить верхнюю цену
        try:
            flightRoutesListPanel = driver.find_element(By.XPATH, "//div[@data-qa-file='FlightRoutesListPanel']")
            # print('flightRoutesListPanel.text:', flightRoutesListPanel.text) # для отладки
            price = flightRoutesListPanel.find_element(By.XPATH, ".//div[@data-qa-file='PricesOffers']").text
            # print('price:', price) # для отладки
            price = price.split('\n')[0].replace(' ', '')
            price = int(price)
        except:
            print('\n    ', sys.exc_info()[1]) # для отладки          
            print('    Билетов нет                   ', end='\r')
        print('    Цена с багажом от', price, '          ')
    
        datePriceS.loc[city + ' ' + date, 'date'] = date
        datePriceS.loc[city + ' ' + date, 'city'] = city
        datePriceS.loc[city + ' ' + date, 'price'] = price
        datePriceS.loc[city + ' ' + date, 'URL'] = url
        if price < priceTarget:
            print(f'    Найдена цена ниже целевой ({priceTarget})', end='\r')
            time.sleep(pause)
    
    return driver, price

def билетов_нет(driver, goF, pause, url):
    print('    Предварительно билетов нет          ')
    driver.close()
    driver = webdriver.Chrome()
    time.sleep(pause * 3)                    
    driver.get(url)
    time.sleep(pause * 3)
    if 'Билетов нет' in driver.find_element("tag name", "body").text:
        print('        Билетов нет          ')
        goF = False

    return driver, goF

def выгрузка_цен(airportS, datePriceS, direction, driver, row):
    goS = True # для взаимодействия с циклом while вне выгрузка_цен()
    try: # обработать ошибку или сигнал сигнал прерывания, поданный на любом этапе сбора данных
        while row <= airportS.index[-1]:
            city = airportS['Город'][row]
            print('\n', city)
            code = airportS['Код'][row]
            
            dateFrom = airportS['Ближайшая дата начала поездки'][row]
            dateFrom = datetime.date(int(str(datetime.date.today()).split('-')[0]), int(dateFrom.split('-')[1]), int(dateFrom.split('-')[2]))
            delta = int(airportS['Максимальный отcтуп'][row])
            
            for dayShift in range(0, delta + 1):
                url = 'https://www.tbank.ru/travel/flights/one-way'
                url += f'/MOW-{code}' if direction == 0 else f'/{code}-MOW'
                date = '-'.join(str(dateFrom + datetime.timedelta(days=dayShift)).split('-')[1:])
                # print('date:', date) # для отладки          
                url += f'/{date}'
                # print("f'/{date}'", f'/{date}') # для отладки          
                url += '/?adults=1&children=0&infants=0&cabin=Y&composite=0'
                if withbaggage: url += '&baggagetype=withbaggage'
                url += f'&transferscount={transferscount}'
                print('                                                  ')
                print('  Захожу по URL-адресу:', url)
                driver.get(url)
                
                firstMoment = datetime.datetime.now()
                while True:
                    try:
                        driver.find_element(By.XPATH, "//span[@data-qa-file='LoadingButton' and contains(., 'Найти')]")
                        print('    Кнопка "Найти" появилась               ') #, end='\r'
                        break   
                    except:
                        print('\n    ', sys.exc_info()[1]) # для отладки
                        print('    Жду кнопку "Найти"')
                        time.sleep(pause)
                        lastMoment = datetime.datetime.now()
                        diffetenceSeconds = lastMoment - firstMoment
                        print('    Ожидание составляет:', diffetenceSeconds.seconds, 'сек.') #, end='\r'
    
                if 'Билетов нет' in driver.find_element("tag name", "body").text:
                    driver, goF = билетов_нет(driver, True, pause, url) # выдача функции здеь не нужна, но пусть будет
                else:
                    driver, price = билеты_есть(city, date, datePriceS, driver, pause, priceTarget, url)
    
            # Конец цика for
    
            row += 1
            if len(datePriceS) > 0:
                print('\n    Промежуточный итог:                          ')
                display(datePriceS[datePriceS['price'] > 0].sort_values('price').head(10))
        # Конец тела внешнего цикла
    
        if len(datePriceS) == 0:
            print('    В интересующий период билетов не найдено')
        return datePriceS, goS, row

    except: # обработать ошибку или сигнал прерывания, поданный на любом этапе сбора данных
        print('\n    ', sys.exc_info()[1]) # для отладки
        goS = False
        return datePriceS, goS, row

driver = webdriver.Chrome()

print('Начинаю поиск билетов туда (если конец сезона, то их можно покупать БЛИЖЕ к дате начала поездки)')
airportS = pandas.read_excel(f'Аэропорты {country}.xlsx', sheet_name='Аэропорты Вылет')
display(airportS)
airportS['Ближайшая дата начала поездки'] = airportS['Ближайшая дата начала поездки'].astype(str)
direction = 0

datePriceS_туда = pandas.DataFrame(columns=['price']) # не внёс внутрь функции выгрузка_цен,
    # чтобы при прерывании исполнения функции можно было продолжить с места прерывания
row = airportS.index[0]

%%time
goS = True
while (row < len(airportS.index)) & goS:
    datePriceS_туда, goS, row = выгрузка_цен(airportS, datePriceS_туда, direction, driver, row)
    datePriceS_туда['direction'] = 'Туда'
print('Завершён поиск билетов туда')

print('Начинаю поиск билетов оттуда (если конец сезона, то их можно покупать ЗАРАНЕЕ)')
airportS = pandas.read_excel(f'Аэропорты {country}.xlsx', sheet_name='Аэропорты Прилёт')
display(airportS)
airportS['Ближайшая дата начала поездки'] = airportS['Ближайшая дата начала поездки'].astype(str)
direction = 1

datePriceS_оттуда = pandas.DataFrame(columns=['price']) # не внёс внутрь функции выгрузка_цен,
    # чтобы при прерывании исполнения функции можно было продолжить с места прерывания
row = airportS.index[0]

%%time
goS = True
while (row < len(airportS.index)) & goS:
    datePriceS_оттуда, goS, row = выгрузка_цен(airportS, datePriceS_оттуда, direction, driver, row)
    datePriceS_оттуда['direction'] = 'Оттуда'
print('Завершён поиск билетов оттуда')

datePriceS = pandas.concat([datePriceS_туда[datePriceS_туда['price'] > 0].sort_values('price'),
                            datePriceS_оттуда[datePriceS_оттуда['price'] > 0].sort_values('price', ascending=False)])
display(datePriceS)

datePriceS['Дата парсинга'] = datetime.date.today().strftime('%Y%m%d')
datePriceS.to_excel(f'parsingTTicket {datetime.date.today().strftime('%Y%m%d')} {country}.xlsx')
try: driver.close()
except: print('driver уже закрыт')

input('Скрипт исполнен')
warnings.filterwarnings("ignore")
sys.exit()



# Тестер
for i in driver.find_elements(By.XPATH, "//div[@data-qa-file='FlightRoutesListPanel']"): # and contains(., 'Дешевле')
    print(i.text)

city = airportS['Город'][row]
date = '04-21'
datePriceS = datePriceS_туда.copy()
url = 'https://www.tbank.ru/travel/flights/one-way/MOW-FJR/04-21/?adults=1&children=0&infants=0&cabin=Y&composite=0&baggagetype=withbaggage&transferscount=1'
driver.get(url)
if 'Билетов нет' in driver.find_element("tag name", "body").text: driver, goF = билетов_нет(driver, True, pause, url)
else:
    driver, price = билеты_есть(city, date, datePriceS, driver, pause, priceTarget, url)



# Авторский модуль для для выяснения, какие инструменты (акции, облигации и т.д.) есть в портфеле, на основе брокерских отчётов

# 0. Активировать требуемые для работы скрипта модули и пакеты 
# sys & subprocess -- эти пакеты должны быть предустанавлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
while True:
    try:
        from datetime import date
        from randan.tools import coLabAdaptor # авторский модуль для адаптации текущего скрипта к файловой системе CoLab
        from randan.trading import bondsFeaturesProcessor # авторский модуль для гармонизации и обработки характеристик облигаций
        import os, pandas, re, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[0]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

coLabFolder = coLabAdaptor.coLabAdaptor()

# 1.0 Авторские функции
# 1.0.0 поиска в брокерском отчёте строчек, ограничивающих интересующий раздел
def boundColibrator(assetS, bound, col, name):
    if len(bound) == 1:
        bound = bound[0]
    elif len(bound) > 1:
        # print('Найдено строк:', len(bound)) # для оталдки
        # display(assetS.loc[bound, col]) # для оталдки
        bound = assetS.loc[bound, col][
            assetS.loc[bound, col].dropna().astype(str).str.contains(name)
                ].index[0]
    return bound

# 1.0.1 организации обработки отчётов каждого брокера за интересующий период
def brokerReportsProcessor(broker, fileNameS, path, period, slash):
    assetS = pandas.DataFrame()
    # print('fileNameS:', fileNameS) # для отладки
    for fileName in fileNameS:
        # print('fileName:', fileName) # для отладки
        assetS_additional = pandas.read_excel(path + broker + slash + 'Отчёты' + slash + fileName, header=None)
        fileNameSpecification = fileName.replace('.xlsx', '').replace(str(period), '')
        # print('fileNameSpecification:', fileNameSpecification) # для отладки
        print(f"Распарсиваю отчёт брокера{'' if fileNameSpecification == '' else ' (' + fileNameSpecification + ')'}", broker, 'за', str(period)) #, end='\r'
        if broker == 'ВТБ': assetS_additional = ВТБ(assetS_additional)
        if broker == 'Тинькофф': assetS_additional = Тинькофф(assetS_additional)
        if broker == 'УралСиб': assetS_additional = УралСиб(assetS_additional)
        assetS_additional.loc[:, 'Брокер'] = broker + fileNameSpecification
        assetS = pandas.concat([assetS, assetS_additional], ignore_index=True)
        # display(assetS['Брокер']) # для отладки
        # display(assetS) # для отладки
    return assetS

# 1.0.2 поиска в брокерском отчёте столбцов, содержащих ключевые данные
def columnFinder(assetS, name):
    # Найти первый слева столбец, содержащий искомое слово(сочетание)
    # print('assetS.columns:', assetS.columns) # для оталдки
    for col in assetS.columns:
        if sum(assetS.loc[:, col].dropna().astype(str).str.contains(name)) > 0:
            # print('col:', col) # для оталдки
            break
    return col

# 1.0.3 поиска в брокерском отчёте имён столбцов, содержащих ключевые данные
def columnNameFinder(assetS, name):
    # Найти первый слева столбец, содержащий искомое слово(сочетание)
    for col in assetS.columns:
        if name in col:
            break
    return col

# 1.0.4 поиска в директориях брокеров отчётов за интересующий период
def reportSearch(broker, path, period, slash):
    fileNameS = []
    goC = True
    while goC:
        for fileName in os.listdir(path + broker + slash + 'Отчёты'):
            if str(period) in fileName:
                # print(f"Нашёл файл '{fileName}'") # для оталдки
                fileNameS.append(fileName)
        if len(fileNameS) > 0:
            goC = False
            return fileNameS, period
            break
        if goC:
            print(f"НЕ нашёл файл за период '{period}'; перехожу к предыдущему периоду")
            period = period - 1 if str(period)[-2:] != '01' else period - 89 # на случай января
            if period == 202001:
                print(f"Безуспешно добрался до периода '{period}'; завершаю поиск и исполнение текущей функции")
                print('Сейчас появится надпись: "An exception has occurred, use %tb to see the full traceback.\nSystemExit" -- так и должно быть')
                input()
                sys.exit()
                break

# 1.0.5 поиска в брокерском отчёте фрагментов, соответствующих интересующим разделам
def sectionFinder(assetS, name, nameNext):
    col = columnFinder(assetS, name)
    upper_bound = assetS[assetS[col].notna() & assetS[col].str.contains(name)].index
    upper_bound = boundColibrator(assetS, upper_bound, col, name)

    lower_bound = assetS[assetS[col].notna()].index[-1]
    if nameNext != '':
        lower_bound = assetS[assetS[col].notna() & assetS[col].str.contains(nameNext)].index
        lower_bound = boundColibrator(assetS, lower_bound, col, name)
    return assetS, lower_bound, upper_bound

# 1.1 Авторские функции-адаптеры по брокерам
def ВТБ(assetS):
    # display(assetS) # для отладки
    # Найти срез датафрейма, соответствующий искомому разделу (раздел предполагает свои наименования столбцов)
    name = 'т об остатках ценных бумаг' # вместо "Отчет", чтобы избежать чередования букв е и ё
    nameNext = ''
    # print('\nИскомый раздел:', name) # для оталдки
    # if nameNext != '': print('Раздел, следующий за искомым:', nameNext) # для оталдки
    assetS, lower_bound, upper_bound = sectionFinder(assetS, name, nameNext)
    colS = list(assetS.loc[upper_bound + 1, :].astype(str)) # поскольку раздел предполагает свои наименования столбцов
    assetS = assetS.loc[upper_bound + 2: lower_bound - 1, :]

    # Найти срез датафрейма, соответствующий искомому подразделу (подраздел НЕ предполагает свои наименования столбцов)
    name = 'ОБЛИГАЦИЯ'
    nameNext = 'ИТОГО:' # Слово(сочетание), следующие за искомым подразделом, должно располагаться в том же столбце
    # print('\nИскомый подраздел:', name) # для оталдки
    # if nameNext != '': print('Слово(сочетание), следующие за искомым подразделом:', nameNext) # для оталдки
    assetS, lower_bound, upper_bound = sectionFinder(assetS, name, nameNext)
    assetS = assetS.loc[upper_bound + 1: lower_bound - 1, :]
    assetS.columns = colS

    # Найти в этом срезе столбцы с ISIN и с позицией
    name = 'ISIN'
    ISIN_clmn = columnNameFinder(assetS, name)
    # print('\nИмя столбца с ISIN:', ISIN_clmn) # для оталдки
    name = 'Плановый исходящий остаток'
    pose_clmn = columnNameFinder(assetS, name)
    # print('\nИмя столбца с позицией:', pose_clmn) # для оталдки
    assetS = assetS[[ISIN_clmn, pose_clmn]]
    assetS.columns = ['ISIN', 'Лотов']

    assetS['ISIN'] = assetS['ISIN'].str.split(',').str[-1]
    assetS = assetS[assetS['Лотов'].notna()] # избавиться от заголовка подтаблицы "ПАИ"
    assetS['Лотов'] = assetS['Лотов'].astype(str).str.split('.').str[0].str.replace(',', '').astype(int)
    # display(assetS) # для отладки
    return assetS

def Тинькофф(assetS):
    # display(assetS) # для отладки
    # Раздел "3.1 Движение по ценным бумагам инвестора"
    # Найти столбец с указанием раздела
    for i in range(len(assetS.columns)):
        if sum(assetS.iloc[:, i].dropna().astype(str).str.contains('Движение ')) > 0:
            break
    boundS = assetS[assetS[i].notna() & assetS[i].str.contains('Движение ')].index
    # print(boundS) # для отладки
    upper_bound = boundS[0]
    lower_bound = boundS[-1]
    assetS = assetS.loc[upper_bound + 1: lower_bound - 1, :]
    # display(assetS) # для отладки

    # Найти столбец с ISIN
    for i in range(len(assetS.columns)):
        if len(assetS.iloc[:, i].dropna()) > 0:
            if sum(assetS.iloc[:, i].dropna().str.contains('RU000')) > 0:
                break
    ISIN_clmn = i # имя столбца с ISIN
    # print(ISIN_clmn)
    bond_indeseS = assetS[assetS[ISIN_clmn].notna() & assetS[ISIN_clmn].str.contains('RU000')].index # строки с бондами
    # print(bond_indeseS) # для отладки
    assetS = assetS.loc[bond_indeseS, :]
    # display(assetS) # для отладки
    clmnS = assetS.loc[bond_indeseS[0], :] # непустые ячейки первой строки с бондами
    # display(clmnS.dropna())
    pose_clmn = clmnS.dropna().index[-1] # имя столбца с Позицией
    # print(pose_clmn) # для отладки
    assetS = assetS[[ISIN_clmn, pose_clmn]]
    assetS.columns = ['ISIN', 'Лотов']
    assetS = assetS[assetS['Лотов'].notna()] # избавиться от заголовка подтаблицы "ПАИ"
    assetS['Лотов'] = assetS['Лотов'].astype(str).str.split('.').str[0].str.replace(',', '').astype(int)
    # display(assetS) # для отладки
    return assetS

def УралСиб(assetS):
    # display(assetS) # для отладки
    assetS = assetS[[3, 10]]
    ISIN_clmn = 3
    bond_indeseS = assetS[assetS[ISIN_clmn].notna() & assetS[ISIN_clmn].str.contains('RU000')].index # строки с акциями и бондами
    # print('bond_indeseS:', bond_indeseS) # для отладки
    assetS = assetS.loc[bond_indeseS, :]
    assetS.columns = ['ISIN', 'Лотов']
    # display(assetS) # для отладки
    return assetS

# 2. Авторская функция исполнения скрипта
def getAssets(
              brokerS=['ВТБ', 'Тинькофф', 'УралСиб'],
              path=coLabFolder,
              returnDfs=False
              ):
    """
    Функция для выяснения, какие облигации есть в портфеле, на основе брокерских отчётов

    Parameters
    ----------
     brokersS : list -- список Ваших брокеров, например, такой: ['ВТБ', 'Тинькофф', 'УралСиб'] ; названия этих трёх брокеров следует писать именно так; для других брокеров функций-адаптеров пока нет.
         path : str -- путь к директории, включая её имя, в которой будут искаться файлы и куда будут сохраняться; по умолчанию, не в CoLab поиск и сохранение происходят в директории, в которой вызывается текущая функция, а в CoLab в директории Colab Notebooks

    returnDfs : bool -- в случае True функция возвращает итоговые датафрейм bondS
    """
        
# 2.0 Настройки
    print('Работаю с месячным периодом; желательно, чтобы все обрабатываемые брокерские отчёты относились к одному периоду')
    period = date.today().strftime("%Y%m%d")[:4] + date.today().strftime("%Y%m%d")[4:6]
    period = int(period) - 89 if period[-2:] == '01' else int(period) - 1 # на случай января
    # print('Целевой период:', period) # для отладки
    
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    if path == None: path = ''
    else: path += slash

    warnings.filterwarnings("ignore")

# 2.1 Выяснение, какие облигации есть в портфеле, на основе брокерских отчётов
    assetS = pandas.DataFrame()
    for broker in brokerS:
        fileNameS, period = reportSearch(broker, path, period, slash)
        # print('broker:', broker) # для отладки        
        assetS_additional = brokerReportsProcessor(broker, fileNameS, path, period, slash)
        # display(assetS_additional) # для отладки
        assetS = pandas.concat([assetS, assetS_additional], ignore_index=True)
    # display(assetS) # для отладки

    # Удалить лишние пробелы во всём накопленном датафрейме
    assetS['ISIN'] = assetS['ISIN'].apply(lambda text: re.sub(r' +', r'', text))
    # display(assetS) # для отладки

# 2.3 Облигации по ключу: ISIN
    assetS_byISIN = pandas.DataFrame(columns=brokerS)
    for isin in assetS['ISIN'].unique():
        # Для каждого брокера создать столбец  
        brokerS_byISIN = assetS[assetS['ISIN'] == isin]
        # display(brokerS_byISIN) # для отладки
        for broker in brokerS_byISIN['Брокер']: # не все brokerS , т.к. конкретный ISIN может встретиться не у всех брокеров
            # display(brokerS_byISIN[brokerS_byISIN['Брокер'] == broker]) # для отладки            
            assetS_byISIN.loc[isin, broker] = brokerS_byISIN.loc[brokerS_byISIN[brokerS_byISIN['Брокер'] == broker].index[0], 'Лотов']
    # display(assetS_byISIN) # для отладки
    assetS_byISIN['Лотов'] = assetS_byISIN.T.sum()
    assetS_byISIN['ISIN'] = assetS_byISIN.index
    assetS_byISIN = assetS_byISIN.reset_index(drop=True)
    # display('assetS_byISIN:', assetS_byISIN) # для отладки

    if returnDfs: return assetS, assetS_byISIN



#!/usr/bin/env python
# coding: utf-8

'''
(EN) A proprietary module that simplifies and manages the web scraping workflow
(RU) Авторский модуль для упрощения скрапинга
'''

def argument_key_comparison(argument, key, params):
    if (key in params.keys()) & (argument != None):
        if params[key] != argument:
            print(f'!!   Вы подали {key} и как аргумент, и через словарь params , причём Вы подали разные значения туда и туда; будет использовано значение, поданное в params !!\n')
            argument = params[key]
    elif (key in params.keys()) & (argument == None): argument = params[key]
    elif (key not in params.keys()) & (argument != None): pass # отдельный аргумент определён, поэтому запрос к пользователю не поступит
    else: pass # НИ ключ params , НИ отдельный аргумент НЕ определены, поэтому запрос к пользователю поступит
    return argument



#!/usr/bin/env python
# coding: utf-8

'''
(EN) A module for topic modelling
(RU) Авторский модуль для тематического моделирования
'''

# 0 Активировать требуемые для работы скрипта модули и пакеты + пререквизиты

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from io import BytesIO
        from openpyxl import Workbook
        from openpyxl.drawing.image import Image
        from openpyxl.utils.dataframe import dataframe_to_rows
        from randan import descriptive_statistics, dimension_reduction
        from tqdm import tqdm
        import matplotlib.pyplot as plt, os, pandas, time, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[1]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break
tqdm.pandas() # для визуализации прогресса функций, применяемых к датафреймам

# 1 Авторские функции для
# 1.0 оформления токенов на полюсах топиков
def summaryPole(loadingsThreshold, minusPlus, tokensLimit, topicDocS, topicLoadingS, topicName):
    summaryPole = pandas.DataFrame()
    if len(topicLoadingS[topicLoadingS[topicName] * minusPlus > loadingsThreshold].sort_values(topicName, ascending=False).head(tokensLimit)) > 0:
        summaryPole = topicDocS[topicDocS[topicName] * minusPlus > 0].round(3)
        summaryPole.loc[:, 'Топик'] = topicName
        summaryPole.loc[:, 'Токены'] = ', '.join(list(topicLoadingS[topicLoadingS[topicName] * minusPlus > loadingsThreshold].sort_values(topicName, ascending=False).head(tokensLimit).index))
        summaryPole.loc[:, 'Усреднённая связь токена с топиком'] = round(topicLoadingS[topicLoadingS[topicName] * minusPlus > loadingsThreshold].sort_values(topicName, ascending=False).head(tokensLimit).mean()[topicName], 3)
        summaryPole.loc[:, 'Релевантность теме исследования'] = ''
        summaryPole.loc[:, 'Интерпретация топика'] = ''
    return summaryPole

# 1.1 описания каждого топика через его полюса и формирующие их токены и релевантные фрагменты располагаемых на них документов
def snippetByDoc(df, loadingsThreshold, pole, poleDocsIndeceS, poleTokenS, supplementarieS):
    print(f'{pole.capitalize()}ый полюс топика')
    if poleTokenS == []:
        print(f'Величина loadings токенов {pole}ого полюса не достигает заданного порога |{round(loadingsThreshold, 2)}|.'
              , f'Поэтому {pole}ый полюс НЕ выражен и НЕ требует интерпретации')
        docs_snippetS = pandas.DataFrame()
    else:
        print(f'{pole.capitalize()}ый полюс сформирован токен[ом, ами]'
              , str(poleTokenS).replace('[', '').replace(']', ''), '-- по величине вклада')
        print(f'На {pole}ом полюсе расположен[ы] документ[ы]', str(poleDocsIndeceS).replace('[', '').replace(']', ''), '-- по близости к полюсу')
        docs_snippetS = pandas.DataFrame()
        for docIndex in poleDocsIndeceS:
            print(f'Посмотрите на фрагменты документа {docIndex}, содержащие указанные выше токены и их окружение.'
                  , 'Документ:', docIndex)
            row = 0
            docSnippetS = pandas.DataFrame(columns=['min', 'max', 'token'])
            for token in poleTokenS:
                goC = True
                # print('Токен:', token)
                textFull_lemmatized_list = df['textFull_lemmatized'][docIndex].split()
                textFull_list = df['textFull_simbolsCleaned'][docIndex].split()
                tokenProximity = 7  # лимит на длину окружения интересующего токена влево и вправо в нелемматизиованном документе
                tokenPosition = -1
                tokenPositionIntervalInitial = [0, 0]
                while goC:
                    try:
                        # Выяснить номер интересующего токена в списке всех токенов лемматизиованного документа
                        tokenPosition = textFull_lemmatized_list.index(token, tokenPosition + 1)

                        tokenPositionInterval = [tokenPosition - tokenProximity, tokenPosition + tokenProximity]
                        # print('tokenPositionIntervalInitial', tokenPositionIntervalInitial)
                        # print('tokenPositionInterval', tokenPositionInterval)

                        if tokenPositionIntervalInitial[-1] > tokenPositionInterval[0]:
                            # print('tokenPositionIntervalInitial и tokenPositionInterval пересеклись\n'
                            #       , 'Поэтому объединяю их\n'
                            #       , 'И не вывожу промежуточный фрагмент на экран')
                            tokenPositionInterval[0] = tokenPositionIntervalInitial[0]
                            # print('Объединённый tokenPositionInterval', tokenPositionInterval)
                        docSnippetS.loc[row, 'min'] = tokenPositionInterval[0]
                        docSnippetS.loc[row, 'max'] = tokenPositionInterval[-1]
                        docSnippetS.loc[row, 'token'] = token
                        tokenPositionIntervalInitial = tokenPositionInterval.copy()
                        row += 1

                    except ValueError:
                        # print('Поиск в документе', docIndex, 'по токену "', token, '"завершён')
                        goC = False
                # display(docSnippetS)
                docSnippetS = docSnippetS.drop_duplicates('min', keep='last')
                # display(docSnippetS)

            rowInvaderToDropS = []
            tokenReceiverS = poleTokenS.copy()
            for tokenInvader in poleTokenS[:-1]:
                tokenInvaderIndex = docSnippetS[docSnippetS['token'] == tokenInvader].index
                tokenReceiverS.remove(tokenInvader)
                for tokenReceiver in tokenReceiverS:
                    # print('tokenInvader:', tokenInvader, '-> tokenReceiver:', tokenReceiver)
                    tokenReceiverIndex = docSnippetS[docSnippetS['token'] == tokenReceiver].index
                    for rowInvader in tokenInvaderIndex:
                        for rowReceiver in tokenReceiverIndex:
                            # print('Иду по:', rowInvader, rowReceiver)
                            if (docSnippetS['min'][rowInvader] >= docSnippetS['min'][rowReceiver])\
                                    & (docSnippetS['min'][rowInvader] <= docSnippetS['max'][rowReceiver])\
                                | (docSnippetS['max'][rowInvader] >= docSnippetS['min'][rowReceiver])\
                                    & (docSnippetS['max'][rowInvader] <= docSnippetS['max'][rowReceiver])\
                                | (docSnippetS['min'][rowReceiver] >= docSnippetS['min'][rowInvader])\
                                    & (docSnippetS['min'][rowReceiver] <= docSnippetS['max'][rowInvader])\
                                | (docSnippetS['max'][rowReceiver] >= docSnippetS['min'][rowInvader])\
                                    & (docSnippetS['max'][rowReceiver] <= docSnippetS['max'][rowInvader]):
                                # print(rowInvader, rowReceiver)
                                docSnippetS.loc[rowReceiver, 'min'] = min(docSnippetS['min'][rowInvader], docSnippetS['min'][rowReceiver])
                                docSnippetS.loc[rowReceiver, 'max'] = max(docSnippetS['max'][rowInvader], docSnippetS['max'][rowReceiver])
                                docSnippetS.loc[rowReceiver, 'token'] += ' ' + tokenInvader
                                rowInvaderToDropS.append(rowInvader)
            docSnippetS = docSnippetS.drop(rowInvaderToDropS)

            for row in docSnippetS.index:
                # По границам интервала вывести окружение интересующего токена в нелемматизиованном документе
                docSnippetS.loc[row, 'pole'] = pole.capitalize() + 'ый'
                docSnippetS.loc[row, 'textSnippet'] = '..' + ' '.join(textFull_list[docSnippetS['min'][row]: docSnippetS['max'][row]]) + '..'
                if supplementarieS != None:
                    for supplementary in supplementarieS:
                        docSnippetS.loc[row, supplementary] = df[supplementary][docIndex]

                print(docSnippetS['textSnippet'][row])
            docs_snippetS = pandas.concat([docs_snippetS, docSnippetS])
        print('\n')
    return docs_snippetS

def randanTopic(dfIn, matrix_df, docsLimit=5, loadingsThreshold=0.5, returnDfs=False, rowsNumerator=None, supplementarieS=None, textFull_lemmatized='textFull_lemmatized', textFull_simbolsCleaned='textFull_simbolsCleaned', tokensLimit=10, topicsCount=None):
    '''    Метод тематического моделирования randanTopic основан на методе главных компонент (по-английски: principal components analisys, PCA). То есть он НЕ использует нейросети и embeddings, а работает с "мешком слов" (по-английски: bag of words, BoW). Поэтому важно качественно подготовить "мешок слов" через очистку исходного текста от лишних символов, лемматизацию и удаление стоп-слов. Эти три этапа предобработки исходного текста, а также (при необходимости) автокоррекция грамматических ошибок могут быть выполнены функциями из скрипта textPreprocessor пакета randan https://github.com/RandanCSS/randan/blob/master/randan/tools/textPreprocessor.py . Причём для удобства последующей интерпретации рекомендую результаты очистки от лишних символов и результаты лемматизации сохранить в отдельные столбцы: textFull_simbolsCleaned и textFull_lemmatized соответственно.
        Если для последующей интерпретации Вам пригодятся дополнительные столбцы (скажем, заголовок, дата и т.п.), впишите их в формате списка текстовых объектов в аргумент supplementaries= текущей функции.
    
    Parameters
    --------------------
                   dfIn : DataFrame -- таблица с исходными данными (текстами и, при ихналичии, вспомогательными переменными)
              matrix_df : DataFrame -- мешок слов
      loadingsThreshold : float -- порог (по модулю) усреднённой связи токена с топиком; нужен для отбора токенов, косвенно лимитирует число токенов при интерпретации         
              docsLimit : int -- лимит на число документов на полюсе топика; нужен для отбора документов            
              returnDfs : bool -- в случае True функция возвращает датафреймы с фрагментами документов docs_snippetS и реквизитами документов и вспомогательными переменными scoreS
          rowsNumerator : str -- имя столбца, значения которого будут использоваться в качестве обозначения строк итоговых таблиц
        supplementarieS : list -- дополнительные столбцы для последующей интерпретации
    textFull_lemmatized : str -- имя столбца с текстом, прошедшем лемматизацию, но из которого НЕ удалены стоп-слова
textFull_simbolsCleaned : str -- имя столбца с текстом, прошедшем удаление лишних символов, но НЕ прошедший лемматизацию и из которого НЕ удалены стоп-слова
            tokensLimit : int -- лимит на число токенов на полюсе топика; нужен для отбора токенов           
            topicsCount : int -- частота самого высокочастотного слова из тех, которые предложены автокорректором в качестве правильного варианта; этот аргумент необходим для корректной работы аргумента userWordS'''
    if (returnDfs == False) & (rowsNumerator == None) & (supplementarieS == None) & (topicsCount == None):
        # print('Пользователь не подал аргументы') # для отладки
        expiriencedMode = False
    else:
        expiriencedMode = True

    if expiriencedMode == False:
        print(
'''    Метод тематического моделирования randanTopic основан на методе главных компонент (по-английски: principal components analisys, PCA). То есть он НЕ использует нейросети и embeddings, а работает с "мешком слов" (по-английски: bag of words, BoW). Поэтому важно качественно подготовить "мешок слов" через очистку исходного текста от лишних символов, лемматизацию и удаление стоп-слов. Эти три этапа предобработки исходного текста, а также (при необходимости) автокоррекция грамматических ошибок могут быть выполнены функциями из скрипта textPreprocessor пакета randan https://github.com/RandanCSS/randan/blob/master/randan/tools/textPreprocessor.py . Причём для удобства последующей интерпретации рекомендую результаты очистки от лишних символов и результаты лемматизации сохранить в отдельные столбцы: textFull_simbolsCleaned и textFull_lemmatized соответственно.
--- После прочтения этой инструкции нажмите Enter'''
              )
        input()

    if 'textFull_lemmatized' not in dfIn.columns:
        print(
f'''В поданном Вами датафрейме отсутствует столбец textFull_lemmatized . Какой столбец в поданном Вами датафрейме содержит лемматизированные тексты (из которого НЕ удалены стоп-слова)?'''
              )
        while True:
            print('--- Впишите, пожалуйста, его название')
            textFull_lemmatized = input()
            if textFull_lemmatized in dfIn.columns: break
            else:
                print('--- Вы вписали что-то не то')

    if 'textFull_simbolsCleaned' not in dfIn.columns:
        print(
f'''В поданном Вами датафрейме отсутствует столбец textFull_simbolsCleaned . Какой столбец в поданном Вами датафрейме содержит тексты, из которых удалены лишние символы, но НЕ лемматизированные?'''
              )
        while True:
            print('--- Впишите, пожалуйста, его название')
            textFull_lemmatized = input()
            if textFull_lemmatized in dfIn.columns: break
            else:
                print('--- Вы вписали что-то не то')  
                    
# 0. Определение желаемого числа топиков
    if topicsCount == None:
        print(
'''Вы уже знаете, сколько топиков хотите, или предпочитаете посмотреть на возможные ориентиры числа топиков, чтобы затем выбрать их число?
--- Если знаете, то введите желаемое число топиков и нажмите Enter
--- Если предпочитаете посмотреть, то просто нажмите Enter -- по умолчанию будут выведены ориентиры для числа топиков не более 30,
поскольку большее число топиков (а) вряд ли захотите проинтерпретировать и (б) расчёт ориентиров для них может занять слишком продолжительное время'''
              )
        while True:
            topicsCount = input()
            try:
                topicsCount = None if topicsCount == '' else int(topicsCount)
                break
            except ValueError:
                print('--- Вы вписали что-то не то'
                      , '\n--- Впишите, пожалуйста, целое число и нажмите Enter: ')

        if topicsCount == None:
            print(
'''Ориентиры для выбора числа топиков: (а) описательная статистика частотности токенов во всём корпусе и в отдельных документах, (б) приемлемый уровнь объяснительной способности модели, критерий Кайзера и скачки объяснительной способности модели (диаграмма "каменистой осыпи"), (в) Ваша субъективная готовность проинтерпретировать не более стольки-то топиков.'''
                  )
# 1.0. Первичный анализ частот токенов в корпусе документов в целом и по каждому документу
# Посчитать сумму по ТОКЕНАМ
            встречаемостьТокенов = pandas.DataFrame(matrix_df.sum(), columns=['Встречаемость токенов'])
            # display(встречаемостьТокенов)
# Посчитать сумму по ДОКУМЕНТАМ
            токеныДокумент = pandas.DataFrame(matrix_df.T.sum(), columns=['Наполненность токенами документов'])
            # display(токеныДокумент)
# 1.1. Описательная статистика для интервальной переменной встречаемостьТокенов
            descriptive_statistics.ScaleStatistics(встречаемостьТокенов[['Встречаемость токенов']])
    # .. и для интервальной переменной токеныДокумент
            descriptive_statistics.ScaleStatistics(токеныДокумент[['Наполненность токенами документов']])

# 2. Пробный запуск (без вращения, но с выводом результатов), чтобы посмотреть `Explained variance`
            dimension_reduction.PCA(n_components='Kaiser' if len(matrix_df.columns) < 30 else 30).fit(matrix_df)
            print('\n--- Введите желаемое число топиков и нажмите Enter')
            while True:
                topicsCount = input()
                try:
                    topicsCount = int(topicsCount)
                    break
                except ValueError:
                    print('--- Вы вписали что-то не то'
                          , '\n--- Впишите, пожалуйста, целое число и нажмите Enter: ')

# 3. Итоговый запуск (с вращением, но с без вывода части результатов)
# Настроить класс PCA: предупредить, сколько будет топиков и какое будет вращение
    if topicsCount > len(matrix_df.columns):
        print('Число топиков принудительно снижено до', len(matrix_df.columns), ', поскольку значение topicsCount, равное', topicsCount, ', слишком велико для располагаемых данных.')
        topicsCount = len(matrix_df.columns)
    pca = dimension_reduction.PCA(n_components=topicsCount, rotation='varimax')
# Подать токены в настроенный класс PCA
    pca = pca.fit(matrix_df, show_results=False)

# 4. Четыре датафрейма, ключевых для оформления и интерпретации результатов тематического моделирования 
    component_loadings_rotated = pca.component_loadings_rotated
    # display(component_loadings_rotated) # для отладки
    topicNameS = component_loadings_rotated.columns

# Матрица "документы-топики" (и величины в ячейках матрицы названы так же)
    df = dfIn.copy()
    scoreS = pca.transform(matrix_df)
    scoreTextS = scoreS.copy()
    
    if rowsNumerator != None:
        if (sum(df[rowsNumerator].isna()) == 0) & (len(list(set(df[rowsNumerator].tolist()))) == len(df)): df.index = df[rowsNumerator] # значения назначенного пользователем столбца будут использоваться в качестве обозначения строк итоговых таблиц, только если эти значения без NaN и без дубликатов
        else: print(
'''
Назначенный Вами столбец для обозначения строк итоговых таблиц содержит NaN или дубликаты, поэтому не будет использоваться.
'''
                    )
    else: # в противном случае останутся исходные обозначения строк, но если эти значения без дубликатов -- иначе reset_index
        if len(list(set(list(df.index)))) != len(df): df.reset_index(drop=True)
        print(
'''
Cреди обозначений строк исходной таблицы есть дубликаты, поэтому эти обозначения заменяю на сквозную нумерацию.
'''
              )
    scoreS.index = df.index
    scoreTextS.index = df.index

    if supplementarieS != None: scoreTextS = pandas.concat([scoreTextS, df[supplementarieS]], axis=1)
    
    # print('df.columns :', df.columns) # для отладки
    
    display(scoreTextS.head())
    print('Число столбцов:', scoreTextS.shape[1], ', число строк', scoreTextS.shape[0])

# 5. Цикл для прохода по всем топикам
    poleS = ['лев', 'прав']
    
    summary = pandas.DataFrame()
    docs_snippetS = pandas.DataFrame()
    errorS = []

    wb = Workbook()
    summary_ws = wb.active
    summary_ws.title = "Сводка"

    for topicName in topicNameS:
    # for topicName in topicNameS[:1]: # для отладки
        print('\n\nTopic', topicName)

    # Документы-топики
        topicScoreS = scoreS[topicName]
    # Токены-топики
        topicLoadingS = component_loadings_rotated[[topicName]]
        # display(topicLoadingS) # для отладки
    # Максимум (по модулю) нагрузки внутри компоненты
        maxLoading = topicLoadingS[topicName].abs().max()
        # print('maxLoading:', maxLoading) # для отладки
        if loadingsThreshold > maxLoading:
            print('Порог (по модулю) усреднённой связи токена с топиком снижен до', round(maxLoading, 2), ', поскольку значение loadingsThreshold, равное', round(loadingsThreshold, 2), ', слишком велико для располагаемых данных.')
            loadingsThreshold = maxLoading * 0.99 # не выше минимума из максимальных по модулю loadings среди компонент

        plt.figure(figsize=(9, 9))
# Распределение документов по осям
    # Гистограмма
        plt.subplot(3, 2, 1) # три строки, три столбца, первое место
        plt.hist(topicScoreS.dropna(), color='grey')
        plt.title(f"Docs Distribution by Scores\nin {topicName}")
        plt.xlabel('Scores')
        plt.ylabel('Frequency');
    # Боксплот
        plt.subplot(3, 2, 2) # три строки, три столбца, третье место
        plt.boxplot(topicScoreS.dropna())
        plt.title(f"Docs Distribution by Scores\nin {topicName}")
        plt.xticks([])
        plt.ylabel('Scores')
        # plt.show()
# Распределение токенов по осям
    # Гистограмма
        plt.subplot(3, 2, 3) # три строки, три столбца, седьмое место
        plt.hist(topicLoadingS.dropna(), color='grey')
        plt.title(f"Tokens Distribution by Loadings\nin {topicName}")
        plt.xlabel('Loadings')
        plt.ylabel('Frequency');
    # Боксплот
        plt.subplot(3, 2, 4) # три строки, три столбца, девятое место
        plt.boxplot(topicLoadingS.dropna())
        plt.title(f"Tokens Distribution by Loadings\nin {topicName}")
        plt.xticks([])
        plt.ylabel('Loadings')
        
        plt.tight_layout()
        img_data = BytesIO()
        plt.savefig(img_data, format='png')
        plt.show() # обязательно после savefig

    # Полярные документы
        topicDocS = pandas.concat([scoreS.sort_values(topicName)[[topicName]].head(docsLimit), scoreS.sort_values(topicName)[[topicName]].tail(docsLimit)])
    # Полярные токены
        topicTokenS = pandas.concat([topicLoadingS[topicLoadingS[topicName] < -loadingsThreshold].sort_values(topicName).head(tokensLimit)
                                     , topicLoadingS[topicLoadingS[topicName] > loadingsThreshold].sort_values(topicName).tail(tokensLimit)])
        print('Токены на полюсах топика', topicName)
        topicTokenS = topicTokenS.dropna()
        display(topicTokenS)
        
    # Обработка полярных документов и токенов; запись в датафрейм
        summaryMinus = summaryPole(loadingsThreshold, -1, tokensLimit, topicDocS, topicLoadingS, topicName)
        summaryPlus = summaryPole(loadingsThreshold, 1, tokensLimit, topicDocS, topicLoadingS, topicName)
        summary_additional = pandas.concat([summaryMinus, summaryPlus])
        summary = pandas.concat([summary, summary_additional])
        
    # Описание каждого топика через его полюса и формирующие их токены и релевантные фрагменты располагаемых на них документов
        poleMinusTokenS = list(topicTokenS[topicTokenS[topicName] < 0].sort_values(topicName).index)
        poleMinusDocsIndeceS = list(topicDocS[topicDocS[topicName] < 0].sort_values(topicName).index)
        minusDocs_snippetS = snippetByDoc(df, loadingsThreshold, poleS[0], poleMinusDocsIndeceS, poleMinusTokenS, supplementarieS)
        polePlusTokenS = list(topicTokenS[topicTokenS[topicName] > 0].index)
        polePlusDocsIndeceS = list(topicDocS[topicDocS[topicName] > 0].index)
        plusDocs_snippetS = snippetByDoc(df, loadingsThreshold, poleS[-1], polePlusDocsIndeceS, polePlusTokenS, supplementarieS)
        
        docs_snippetS_additional = pandas.concat([minusDocs_snippetS, plusDocs_snippetS])
        docs_snippetS_additional = docs_snippetS_additional.drop(['min', 'max'], axis=1)
        docs_snippetS_additional = docs_snippetS_additional.reset_index(drop=True)
        docs_snippetS_additional.loc[:, 'Интерпретация топика'] = ''
        docs_snippetS = pandas.concat([docs_snippetS, docs_snippetS_additional])
        # print('\n\n\n')

        ws = wb.create_sheet(title=topicName)
        for r in dataframe_to_rows(docs_snippetS_additional, index=False, header=True):
            ws.append(r)
        img_data.seek(0)
        img = Image(img_data)
        ws.add_image(img, "K1")

    docs_snippetS = docs_snippetS.reset_index(drop=True)

# 6. Если норм, отправить эти документы в Excel
    # Желательно после начала интерпретации назвать Excel по-новому, чтобы он не перезаписался в дальнейшем.
    summaryColumns = ['Топик', 'Токены', 'Усреднённая связь токена с топиком']
    summaryColumns.extend(scoreS.columns)
    summaryColumns.extend(['Релевантность теме исследования', 'Интерпретация топика'])

    for r in dataframe_to_rows(summary[summaryColumns], index=False, header=True):
        summary_ws.append(r)

    fileName = 'Топики'
    attempt = 0
# Чтобы не перезаписать ранее созднный файл, в котором может быть интерпретация
    while os.path.exists(fileName + ' ' + str(attempt) + ".xlsx"):
        attempt += 1

    print(f'''\nCоздаю файл "{fileName + ' ' + str(attempt)}.xlsx"''')
    wb.save(fileName + ' ' + str(attempt) + ".xlsx")

    # # Причём разместить релевантные фрагменты документов на отдельныъх страницах
    # with pandas.ExcelWriter("Топики.xlsx") as writer:
    #     summary[summaryColumns].to_excel(writer, sheet_name="Сводка")
    #     for topicName in scoreS.columns:
    #         docs_snippetS[docs_snippetS['topicName'] == topicName].to_excel(writer, sheet_name=topicName)

    print(
'''
Рекомендации по дальнейшей интерпретации топиков:
1.1. Интерпретация топиков по ключевым токенам: на листе Сводка посмотрите перечень ключевых токенов каждого топика. Если токенов слишком мало или много, перезапустите скрипт, понизив или повысив порог loadingsThreshold соответственно, либо повысив или понизив порог tokensLimit соответственно.
1.2. Попробуйте в одной или нескольких фразах сормулировать смысл каждого топика по его ключевым токенам и внесите сформулированный смысл в столбец Интерпретация топика. Если у топика два полюса, лучше для каждого из них формулировать смысл отдельно.

2.1. Интерпретация топиков по выдержкам из ключевых документов и ключевым токенам: на листе Сводка посмотрите перечень ключевых документов каждого топика. Если документов слишком мало или много, перезапустите скрипт, повысив или понизив порог docsLimit соответственно.
2.2. На листе PC1_vrmx читайте выдержки (с акцентом на соответствующие ключевые токены) и при обнаружении принципиально нового смысла, вносите его в свободный столбец. Нюансы и оттенки уже выписанных смыслов лучше игнорировать, чтобы не "закопаться".
2.3. Повторите процедуру для остальных листов PC..._vrmx. По завершении постарайтесь в одной или нескольких фразах обобщить смыслы для каждого листа (топика) и внесите сформулированные обобщения в столбец Интерпретация топика листа Сводка. Если у топика два полюса, лучше для каждого из них формулировать смысл отдельно.

3.1. Интерпретация топиков по ключевым документам и ключевым токенам с помощью ИИ: на листе Сводка посмотрите перечень ключевых токенов и документов каждого топика. Если их слишком мало или много, выполните пункты 2.1 или 3.1.
3.2. На листе Сводка выберите несколько (можете ориентироваться на диаграмму Docs distribution by scores...) ключевых документов с максимальным значением в столбце PC1_vrmx , подайте их в ИИ с просьбой вычленить основные мысли с привязкой к соответствующим ключевым токенам.
3.3. Ответ ИИ внесите в столбец Интерпретация топика напротив первого топика. Повторите процедуру для остальных топиков.'''
          )
    if returnDfs: return docs_snippetS, scoreS



from itertools import combinations
from matplotlib import pyplot as plt
import os
import numpy as np
import pandas as pd
from factor_analyzer.rotator import Rotator
from sklearn.base import BaseEstimator
import scipy as sp
import statsmodels.formula.api as smf
import time


from IPython.display import display

class CA:
    
    """
    A class for implementing correspondence analysis (CA),
    a method for identifying associations of two categorical
    variables.
    
    Parameters
    ----------
    n_dimensions : int 
        A number of dimensions in solution (equals 2 by default)
    
    Attributes
    ----------
    crosstab : pd.DataFrame
        A contingency table
    correspondence_matrix : pd.DataFrame
        A normalized (divided by number of observations) contingency table 
    row_categories : list
        Names of the categories located at the rows of a crosstab
    column_categories : list
        Names of the categories located at the columns of a crosstab
    row_mass : pd.Series
        Masses of row categories
    column_mass : pd.Series
        Masses of column categories
    inertia_total : float
        Total inertia of a crosstab
    inertia_by_dimensions : pd.DataFrame
        Inertia accounted for
    inertia_of_rows : np.ndarray
        Inertia of row categories
    inertia_of_columns : np.ndarray
        Inertia of column categories
    principal_row_coordinates : np.ndarray
        Principal coordinates of row categories
    principal_column_coordinates : np.ndarray
        Principal coordinates of column categories
    standard_row_coordinates : np.ndarray
        Standard coordinates of row categories
    standard_column_coordinates : np.ndarray
        Standard coordinates of column categories
    contribution_of_rows_to_inertia_of_dimensions : np.ndarray
        Contribution of row categories to inertia of dimensions
    contribution_of_columns_to_inertia_of_dimensions : np.ndarray
        Contribution of column categories to inertia of dimensions
    contribution_of_dimensions_to_inertia_of_rows : np.ndarray
        Contribution of dimensions to inertia of row categories 
    contribution_of_dimensions_to_inertia_of_columns : np.ndarray
        Contribution of dimensions to inertia of column categories
    communalities_rows : np.ndarray
        Analogues of communalities for row categories
    communalities_columns : np.ndarray
        Analogues of communalities for column categories
    component_loadings_rows : pd.DataFrame
        Analogues of component (factor) loadings for row categories
    component_loadings_columns : pd.DataFrame
        Analogues of component (factor) loadings for column categories
    """
    
    def __init__(self, n_dimensions=2):
        self.n_dimensions = n_dimensions    

    def fit(
        self, 
        data, 
        row=None, 
        column=None, 
        data_type='raw',
        show_results=True,
        n_decimals=3,
        plot_dimensions=True
    ):
        
        """
        Fit a model to the given data.
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model. 
            Either the raw data or the contingency table with row categories as index
        row : str 
            [only necessary if raw data are passed]
            Name of a variable of the data to be considered as row values
        column : str 
            [only necessary if raw data are passed]
            Name of a variable of the data to be considered as column values
        data_type : str 
            Type of the data passed, 
            possible values are 'raw' or 'crosstab' 
        show_results : bool 
            Whether to show results of the analysis
        n_decimals : int 
            Number of digits to round results when showing them
        plot_dimensions : bool 
            Whether to vizualize two first dimensions

        Returns
        -------
        self
            The current instance of the CA class
        """
        
        if data_type.lower() == 'raw':
            if len(data.columns) > 2 and (not row and not column):
                raise ValueError("""Number of variables in DataFrame
                exceeds 2. Please specify both row and column arguments
                or filter the necessary variables manually.""")

            elif len(data.columns) < 2:
                raise ValueError("""One or no variables were passed.
                Please add necessary variables to DataFrame.""")

            if row and column:
                self.crosstab = pd.crosstab(data[row], data[column])

            elif row or column:
                raise ValueError("""Please specify both row and column arguments
                or filter the necessary variables manually.""")

            else: 
                self.crosstab = pd.crosstab(data.iloc[:, 0], data.iloc[:, 1])

        elif data_type.lower() == 'crosstab':
            
            self.crosstab = data.copy()
            
        else:
            raise ValueError("""Unknown data type. Possible values: 'raw' or 'crosstab'""")
            
        while len(self.crosstab.loc[self.crosstab.var(axis=1) == 0.0, :]) > 0 or \
        len(self.crosstab.loc[:, self.crosstab.var() == 0.0].T) > 0:
            self.crosstab = self.crosstab.loc[:, self.crosstab.var() > 0.0].copy()
            self.crosstab = self.crosstab.loc[self.crosstab.var(axis=1) > 0.0, :].copy()
            print('One or more categories were removed from analysis as they have zero variance.')
        
        max_inertia = min(len(self.crosstab.columns),
                         len(self.crosstab.index)) - 1
        
        if self.n_dimensions > max_inertia:
            raise ValueError(f"""Too many dimensions requested. The maximum possible number of
            dimensions is {max_inertia}, but {self.n_dimensions} were requested.""")
        
        self.N = self.crosstab.sum().sum()
        self.column_marginals = self.crosstab.sum()
        self.row_marginals = self.crosstab.sum(axis=1)
        self.correspondence_matrix = self.crosstab / self.N
        self.row_categories = list(self.correspondence_matrix.index)
        self.column_categories = list(self.correspondence_matrix.columns)
        
        
        self.column_profiles = self.crosstab / self.column_marginals
        
        try:
            self.column_profiles.columns = self.column_profiles.columns.add_categories('Mean column profile')
            self.column_profiles['Mean column profile'] = self.row_marginals / self.N
        except AttributeError:
            self.column_profiles['Mean column profile'] = self.row_marginals / self.N

        
        self.row_profiles = (self.crosstab.T / self.row_marginals).T
        
        try:
            self.row_profiles.index = self.row_profiles.index.add_categories('Mean row profile')
            self.row_profiles.loc['Mean row profile'] = self.column_marginals / self.N
        except AttributeError:
            self.row_profiles.loc['Mean row profile'] = self.column_marginals / self.N
            

        
        self.row_mass = self.column_profiles['Mean column profile'].copy()
        self.row_mass.name = 'Row mass'
        self.column_mass = self.row_profiles.loc['Mean row profile'].copy()
        self.column_mass.name = 'Column mass'
        
        self.expected_frequencies = pd.DataFrame(np.outer(self.correspondence_matrix.sum(),
                                                          self.correspondence_matrix.sum(axis=1)).T,
                                                 index = self.correspondence_matrix.index,
                                                 columns = self.correspondence_matrix.columns)
        
        self.standardized_residuals = (self.correspondence_matrix                                        - self.expected_frequencies) / np.sqrt(self.expected_frequencies)
        
        self.inertia_total = (self.standardized_residuals ** 2).sum().sum()
        
        self.U, self.singular_values, self.V = np.linalg.svd(self.standardized_residuals, full_matrices=True)
        
        self.inertia_by_dimensions = pd.DataFrame(self.singular_values ** 2,
                                                        columns = ['Inertia'],
                                                        index = [f'Dimension {i+1}' for i in range(len(self.singular_values))])
        self.inertia_by_dimensions['Inertia (%)'] = self.inertia_by_dimensions['Inertia'] / self.inertia_total * 100
        self.inertia_by_dimensions = self.inertia_by_dimensions.iloc[:self.n_dimensions, :]
        self.inertia_by_dimensions.loc['Total'] = self.inertia_by_dimensions.sum()
        
        self.standard_row_coordinates = (np.diag(self.row_mass.values**-0.5) @ self.U)[:, :max_inertia]
        self.standard_column_coordinates = np.diag(self.column_mass.values**-0.5) @ self.V[:max_inertia, :].T
        
        self.principal_row_coordinates = self.standard_row_coordinates @ np.diag(self.singular_values[:max_inertia])
        self.principal_column_coordinates = np.diag(self.column_mass.values**-0.5) @ self.V[:max_inertia, :].T @ np.diag(self.singular_values[:max_inertia])
        
        self.row_inertia_components = np.diag(self.row_mass.values) @ self.principal_row_coordinates**2
        self.column_inertia_components = np.diag(self.column_mass.values) @ self.principal_column_coordinates**2
        
        self.contribution_of_rows_to_inertia_of_dimensions = self.row_inertia_components / self.row_inertia_components.sum(axis=0)[np.newaxis, :]
        self.contribution_of_columns_to_inertia_of_dimensions = self.column_inertia_components / self.column_inertia_components.sum(axis=0)[np.newaxis, :]
        
        self.inertia_of_rows = self.row_inertia_components.sum(axis=1)
        self.inertia_of_columns = self.column_inertia_components.sum(axis=1)
        
        self.contribution_of_dimensions_to_inertia_of_rows = self.row_inertia_components / self.inertia_of_rows[:, np.newaxis]
        self.contribution_of_dimensions_to_inertia_of_columns = self.column_inertia_components / self.inertia_of_columns[:, np.newaxis]
        
        self.communalities_rows = self.contribution_of_dimensions_to_inertia_of_rows[:, :self.n_dimensions].sum(axis=1)
        self.communalities_columns = self.contribution_of_dimensions_to_inertia_of_columns[:, :self.n_dimensions].sum(axis=1)
        
        self.component_loadings_rows = self.contribution_of_dimensions_to_inertia_of_rows**0.5 * np.sign(self.principal_row_coordinates)
        self.component_loadings_rows = pd.DataFrame(self.component_loadings_rows[:, :self.n_dimensions], 
                                                columns = [f'Dimension {i+1}' for i in range(self.n_dimensions)],
                                                index = self.correspondence_matrix.index)
        self.component_loadings_columns = self.contribution_of_dimensions_to_inertia_of_columns**0.5 * np.sign(self.principal_column_coordinates)
        self.component_loadings_columns = pd.DataFrame(self.component_loadings_columns[:, :self.n_dimensions], 
                                                columns = [f'Dimension {i+1}' for i in range(self.n_dimensions)],
                                                index = self.correspondence_matrix.columns)
        
        if show_results:
            self.show_results(n_decimals)
            
        if plot_dimensions:
            print('------------------\n')
            if self.n_dimensions == 1:
                print("The plot couldn't be drawn with n_demnsions set to 1.")
            else:
                print("Dimension's plot")
                self.plot_dimensions()
                if self.n_dimensions > 2:
                    print('Only two first dimensions are shown.')
                    print(f'To get more plots, use [model].plot_dimensions((1, {self.n_dimensions}))')
        
        return self
    
    def show_results(self, n_decimals):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int 
            Number of digits to round results when showing them
        """
        print('\nCA SUMMARY')
        print('------------------\n')
        print('Explained inertia')
        display(self.inertia_by_dimensions.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("attribute .inertia_by_dimensions"))
        print('------------------\n')
        print('Detailed information')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("method .summary()"))
    
    def summary(self, 
                display_component_loadings=True, 
                display_contributions=False,
                display_coordinates=False):
        """
        Return summary of the fitted model
        in terms of mass, inertia, contributions and communalities.
        
        Parameters
        ----------
        display_component_loadings : bool 
            Whether to display component loadings 
            instead of 'raw' contributions of dimensions and points to inertia in the final table
        display_contributions : bool 
            Whether to display 
            'raw' contributions of dimensions and points in the final table
        display_coordinates : bool 
            Whether to display 
            coordinates of points in the final table

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        
        summary_table = pd.DataFrame(index = self.row_categories + self.column_categories)
        summary_table['Mass'] = list(self.row_mass) + list(self.column_mass)
        
        if display_coordinates==True:
            for i in range(self.n_dimensions):
                summary_table[f'Coordinates (dim. {i+1})'] = list(self.principal_row_coordinates[:, i]) + list(self.principal_column_coordinates[:, i])
        
        summary_table['Inertia (%)'] = list(self.inertia_of_rows / self.inertia_total * 100) + list(self.inertia_of_columns / self.inertia_total * 100)
        
        if display_contributions==True:
            for i in range(self.n_dimensions):
                summary_table[f'Contribution of points to inertia of dimensions (dim. {i+1})'] = list(self.contribution_of_rows_to_inertia_of_dimensions[:, i]) + list(self.contribution_of_columns_to_inertia_of_dimensions[:, i])
            
            for i in range(self.n_dimensions):
                summary_table[f'Contribution of dimensions to inertia of points (dim. {i+1})'] = list(self.contribution_of_dimensions_to_inertia_of_rows[:, i]) + list(self.contribution_of_dimensions_to_inertia_of_columns[:, i])
            
        if display_component_loadings==True:
            for i in range(self.n_dimensions):
                summary_table[f'Component loadings of points (dim. {i+1})'] = list(self.component_loadings_rows.iloc[:, i]) + list(self.component_loadings_columns.iloc[:, i])        
         
        summary_table[f'Communalities ({self.n_dimensions} dims)'] = list(self.communalities_rows) + list(self.communalities_columns)
        
        summary_table.fillna(0, inplace=True)
        
        return summary_table
    
    def show_components(self, axis=0, n_indicators=None):
        
        """
        Show component loadings in a more concise way.
        Useful when dealing with big crosstabs 
        (e.g., when CA is used to text data for topic modeling).
        
        Parameters
        ----------
        axis : int or str 
            Categories of which axis to show, 
            possible values are 'row' (0) or 'column' (1)
        n_indicators : int 
            How many indicators to show, if not specified,
            all indicators will be shown

        Returns
        -------
        pd.DataFrame
            A table with indicators and their loadings
        """
        row_categories = [str(category) + '__ROW' for category in self.row_categories]
        column_categories = [str(category) + '__COLUMN' for category in self.column_categories]
        
        if axis == 0 or axis == 'row':
            categories_to_show = row_categories.copy()
            
        elif axis == 1 or axis == 'column':
            categories_to_show = column_categories.copy()
            
        else:
            raise ValueError("Unknown axis. Possible values: 'row' (0) and 'column' (1)")
        
        if not n_indicators:
            n_indicators = len(categories_to_show)
            
        components = pd.DataFrame(index = range(n_indicators))
        n_components = self.n_dimensions
        summary = self.summary().copy()
        summary.index = row_categories + column_categories
        
        for i in range(2, n_components+2):
            positive_loadings_index = list(summary.loc[categories_to_show].iloc[:, i].sort_values(ascending=False)[:n_indicators].index)
            positive_loadings_values = list(summary.loc[categories_to_show].iloc[:, i].sort_values(ascending=False)[:n_indicators])
            positive_loadings = list(zip(positive_loadings_index, positive_loadings_values))
            positive_loadings = [str(loading)+'*'+index for index, loading in positive_loadings]    
            components[f'Component {i-1} (positive loadings)'] = positive_loadings

            negative_loadings_index = list(summary.loc[categories_to_show].iloc[:, i].sort_values(ascending=True)[:n_indicators].index)
            negative_loadings_values = list(summary.loc[categories_to_show].iloc[:, i].sort_values(ascending=True)[:n_indicators])
            negative_loadings = list(zip(negative_loadings_index, negative_loadings_values))
            negative_loadings = [str(loading)+'*'+index for index, loading in negative_loadings]
            components[f'Component {i-1} (negative loadings)'] = negative_loadings
            
        for component in components.columns:
            components[component] = components[component].str.replace('__ROW', '').str.replace('__COLUMN', '')
        
        return components
    
    def plot_dimensions(
        self, 
        dimensions_range=(1, 2), 
        size=100, 
        n_symbols=20, 
        save=False
    ):

        """
        Build plots of categories's coordinates in the obtained dimensions.
        
        Parameters
        ----------
        dimensions_range : tuple 
            Which dimensions to consider,
            e.g. if dimensions_range is (1, 3), it will produce three plots
            (the 1st dimension & the 2nd dimension,
            the 2nd dimension & the 3rd dimension,
            the 1st dimension & the 3rd dimension)
        size : int 
            Size of plots, in percentage (100 corresponds to figsize=(8, 5))
        n_symbols : int 
            How many symbols of labels to display
        save : bool 
            Whether to save plots or not (in the current directory)
        """
        
        if self.n_dimensions == 1:
            raise ValueError("The plot couldn't be drawn with n_demnsions set to 1.")
        
        dimension_low = dimensions_range[0]-1
        dimension_high = dimensions_range[1]-1

        dimension_combinations = list(combinations(range(dimension_low, dimension_high+1), 2))

        n_plots = len(dimension_combinations)
        size = size/100
        for plot in range(n_plots):
            plt.figure(figsize=(8*size, 5*size))
            current_combination_x, current_combination_y = dimension_combinations[plot]
            x1, y1 = self.principal_row_coordinates[:, current_combination_x], self.principal_row_coordinates[:, current_combination_y]
            x2, y2 = self.principal_column_coordinates[:, current_combination_x], self.principal_column_coordinates[:, current_combination_y]
            plt.scatter(x1, y1, c='r')
            plt.scatter(x2, y2, c='gray', marker='s')

            x_delta = 0.005
            y_delta = 0.005

            for i, txt in enumerate(self.row_categories):
                plt.annotate(txt[:n_symbols]+'...', (x1[i]-x_delta, y1[i]+y_delta))

            for i, txt in enumerate(self.column_categories):
                plt.annotate(txt[:n_symbols]+'...', (x2[i]-x_delta, y2[i]+y_delta))

            plt.axhline(0, alpha=0.2, linestyle='--')  #horizontal line
            plt.axvline(0, alpha=0.2, linestyle='--')
            x_min = min(plt.xticks()[0])
            x_max = max(plt.xticks()[0])

            y_min = min(plt.yticks()[0])
            y_max = max(plt.yticks()[0])

            xticks = plt.xticks()[0]

            plt.fill_between(xticks, y_min, 0,
                            where = xticks <= 0,
                            alpha=0.2)
            plt.fill_between(xticks, y_min, 0,
                            where = xticks >= 0,
                            alpha=0.2)
            plt.fill_between(xticks, 0, y_max,
                            where = xticks <= 0,
                            alpha=0.2)
            plt.fill_between(xticks, 0, y_max,
                            where = xticks >= 0,
                            alpha=0.2);
            plt.xlabel(f'Dimension {current_combination_x+1}')
            plt.ylabel(f'Dimension {current_combination_y+1}')
            plt.grid(alpha=0.15)
            if save:
                plt.savefig(f'Plot_{plot}.png')
            plt.show();

# this class is not for direct use
class Rotator():

    """
    The Rotator class takes an (unrotated) component loading matrix and performs one of several rotations.

    Parameters
    ----------
    method : str, optional
        The component rotation method. Options include:
            (a) varimax (orthogonal rotation)
            (b) promax (oblique rotation)
            
            ОСТАЛЬНЫЕ ПОКА НЕ РАБОТАЮТ
            (c) oblimin (oblique rotation)
            (d) oblimax (orthogonal rotation)
            (e) quartimin (oblique rotation)
            (f) quartimax (orthogonal rotation)
            (g) equamax (orthogonal rotation)
        Defaults to 'varimax'.
    normalize : bool or None, optional
        Whether to perform Kaiser normalization and de-normalization prior to and following rotation.
        Used for varimax and promax rotations.
        If None, default for promax is False, and default for varimax is True.
        Defaults to None.
    gamma : int, optional
        The gamma level for the oblimin objective.
        Ignored if the method is not 'oblimin'.
        Defaults to 0.
    kappa : int, optional
        The extention to which raise the components' correlation within 'promax'.
        Numbers generally range form 2 to 4.
        Ignored if the method is not 'equamax'.
        Defaults to 4.
    max_iter : int, optional
        The maximum number of iterations.
        Used for varimax and oblique rotations.
        Defaults to `1000`.
    tol : float, optional
        The convergence threshold.
        Used for varimax and oblique rotations.
        Defaults to `1e-5`.

    Attributes
    ----------
    loadings_ : pandas DataFrame
        The loadings matrix
    rotation_ : numpy array, shape (n_components, n_components)
        The rotation matrix
    psi_ : numpy array or None
        The component correlations
        matrix. This only exists
        if the rotation is oblique.

    Notes
    -----


    References
    ----------
    [1] https://factor-analyzer.readthedocs.io/en/latest/_modules/factor_analyzer/rotator.html

    Examples
    --------
    >>> rotator = RotatorBiggs(ars_list=vars_init_list,PCs_list=PCs_init_list_1,loadings_=loadings_init_df_1)
    >>> rotator.fit_transform()
    DataFrame...
    """

    def __init__(self,vars_list,PCs_list,loadings_,
                 method='varimax',
                 normalize=True,
                 gamma=0,
                 kappa=4,
                 max_iter=500,
                 tol=1e-5):

        self.method = method
        self.normalize = normalize
        self.gamma = gamma
        self.kappa = kappa
        self.max_iter = max_iter
        self.tol = tol
        self.vars_list = vars_list
        self.PCs_list = PCs_list
        self.loadings_ = loadings_
        
        self.rotation_ = None
        self.phi_ = None
        self.X = None

    def _varimax(self, X=None):
        """
        Perform varimax (orthogonal) rotation, with optional Kaiser normalization.

        Parameters
        ----------


        Returns
        -------
        self
        """
        if X is None:
            X = self.loadings_.copy()
        
        n_rows, n_cols = X.shape
        if n_cols < 2:
            return X

        # normalize the loadings matrix using sqrt of the sum of squares (Kaiser)
#         start = time.time()
        if self.normalize:
            normalized_mtx = np.apply_along_axis(lambda x: np.sqrt(np.sum(x**2)),1,X.copy())
            X = (X.T/normalized_mtx).T
#         print(f"varimax-normalized_mtx: {time.time()-start}")

        # initialize the rotation matrix to N x N identity matrix
        rotation_mtx = np.eye(n_cols)

#         start = time.time()
        d = 0
        for _ in range(self.max_iter):

            old_d = d

            # take inner product of loading matrix and rotation matrix
            basis = np.dot(X,rotation_mtx)

            # transform data for singular value decomposition
            transformed = np.dot(X.T,basis**3-(1.0/n_rows)*np.dot(basis,np.diag(np.diag(np.dot(basis.T,basis)))))

            # perform SVD on the transformed matrix
            U,S,V = np.linalg.svd(transformed)
            
            # take inner product of U and V, and sum of S
            rotation_mtx = np.dot(U,V)
            d = np.sum(S)

            # check convergence
            if old_d != 0 and d/old_d < 1+self.tol:
#                 print(f'The rotation converged in {_} iterations')
                break
#             if _== self.max_iter-1:
#                 print(f'The rotation is not converged in {_} iterations')
                
#         print(f"varimax-max_iter: {time.time()-start}")

        # take inner product of loading matrix and rotation matrix
        X = np.dot(X,rotation_mtx)


        # de-normalize the data
#         start = time.time()
        if self.normalize:
            X = X.T*normalized_mtx
        else:
            X = X.T
#         print(f"varimax-DEnormalize: {time.time()-start}")

        # я вписал
        # convert loadings matrix to data frame
        loadings = pd.DataFrame(X.T,columns=self.loadings_.columns,index=self.loadings_.index)
        
        variance = self._get_component_variance(loadings)[0]
        loadings = loadings[variance.sort_values(ascending=False).index]
        
        PCs_vrmx_list = [f'PC{i}_vrmx' for i in range(1, len(self.PCs_list) + 1)]
        loadings.columns = PCs_vrmx_list
        
        return loadings, rotation_mtx

    def _promax(self):
        """
        Perform promax (oblique) rotation, with optional Kaiser normalization.

        Parameters
        ----------


        Returns
        -------
        loadings : numpy array, shape (n_features, n_components)
            The loadings matrix
        rotation_mtx : numpy array, shape (n_components, n_components)
            The rotation matrix
        psi : numpy array or None, shape (n_components, n_components)
            The component correlations
            matrix. This only exists
            if the rotation is oblique.
        """
        X = self.loadings_.copy()
        
        if self.normalize:
            # pre-normalization is done in R's
            # `kaiser()` function when rotate='Promax'.
            array = X.copy()
            h2 = sp.diag(np.dot(array, array.T))
            h2 = np.reshape(h2, (h2.shape[0], 1))
            weights = array / sp.sqrt(h2)

        else:
            weights = X.copy()
        # first get varimax rotation
        X, rotation_mtx = self._varimax(weights)
        Y = X*np.abs(X)**(self.kappa-1)

        # fit linear regression model
        coef = np.dot(np.linalg.inv(np.dot(X.T,X)),np.dot(X.T,Y))

        # calculate diagonal of inverse square
        try:
            diag_inv = sp.diag(sp.linalg.inv(sp.dot(coef.T,coef)))
        except np.linalg.LinAlgError:
            diag_inv = sp.diag(sp.linalg.pinv(sp.dot(coef.T,coef)))

        # transform and calculate inner products
        coef = sp.dot(coef,sp.diag(sp.sqrt(diag_inv)))
        z = sp.dot(X,coef)

        if self.normalize:
            # post-normalization is done in R's
            # `kaiser()` function when rotate='Promax'
            z = z*sp.sqrt(h2)

        rotation_mtx = sp.dot(rotation_mtx,coef)

        coef_inv = np.linalg.inv(coef)
        phi = np.dot(coef_inv, coef_inv.T)

        # я вписал
        # convert loadings matrix to data frame
        loadings = pd.DataFrame(z, 
                                columns=self.loadings_.columns,
                                index=self.loadings_.index)
        
        variance = self._get_component_variance(loadings)[0]
        loadings = loadings[variance.sort_values(ascending=False).index]
        
        PCs_vrmx_list = [f'PC{i}_prmx' for i in range(1, len(self.PCs_list) + 1)]
        loadings.columns = PCs_vrmx_list
        
        return loadings, rotation_mtx, phi
    
    @staticmethod
    def _get_component_variance(loadings):
        """
        A helper method to get the component variances,
        because sometimes we need them even before the model is fitted.

        Parameters
        ----------
        loadings : array-like
            The component loading matrix,
            in whatever state.

        Returns
        -------
        variance : numpy array
            The component variances.
        proportional_variance : numpy array
            The proportional component variances.
        cumulative_variances : numpy array
            The cumulative component variances.
        """
        n_rows = loadings.shape[0]

        # calculate variance
        loadings = loadings**2
        variance = np.sum(loadings,axis=0)

        # calculate proportional variance
        proportional_variance = variance/n_rows

        # calculate cumulative variance
        cumulative_variance = np.cumsum(proportional_variance, axis=0)

        return (variance,
                proportional_variance,
                cumulative_variance)
        
    def fit(self):
        """
        Computes the component rotation.

        Parameters
        ----------


        Returns
        -------
        self
        """
        self.fit_transform(self.X)
        return self
    
    def fit_transform(self):
        """
        Computes the component rotation,
        and returns the new loading matrix.

        Parameters
        ----------

        Returns
        -------
        loadings_ : pandas DataFrame
            The loadings matrix

        Raises
        ------
        ValueError
            If the `method` is not in the list of
            acceptable methods.

        Example
        -------
        >>> rotator = RotatorBiggs(ars_list=vars_init_list,PC_list=PCs_init_list_1,loadings_=loadings_init_df_1)
        >>> rotator.fit_transform()
        DataFrame...
        """
        # default phi to None
        # it will only be calculated
        # for oblique rotations
        phi = None
        method = self.method.lower()
        if method == 'varimax':
            (new_loadings,
             new_rotation_mtx) = self._varimax()

        elif method == 'promax':
            (new_loadings,
             new_rotation_mtx,
             phi) = self._promax()

#         elif method in OBLIQUE_ROTATIONS:
#             (new_loadings,
#              new_rotation_mtx,
#              phi) = self._oblique(X,method)

#         elif method in ORTHOGONAL_ROTATIONS:
#             (new_loadings,
#              new_rotation_mtx) = self._orthogonal(X,method)

        else:
            raise ValueError("The value for `method` must be one of the "
                             "following: {}.".format(', '.join(POSSIBLE_ROTATIONS)))

        (self.loadings_,
         self.rotation_,
         self.phi_) = new_loadings, new_rotation_mtx, phi
        return self.loadings_

class PCA: # первичен
    
    """A class for implementing principal component analysis (PCA).
    
    Parameters
    ----------
    n_components : int or str 
        The exact number of dimensions
        in solution or the criterion for its automatic selection.
        Current possible values: 'kaiser' (defualt),
        which corresponds to Kaiser's criterion
    rotation : None, 'varimax', 'promax', 'natural collinearity' 
        Rotation to perform on factor loadings.
        Currently, only varimax rotation is available
    kappa : int
        Kappa parameter for promax rotation


    Attributes
    ----------
    correlation_matrix : pd.DataFrame
        A correlation matrix
    explained_variance : pd.DataFrame
        A table with eigenvalues and variance accounted for
    explained_variance_total : float
        Total percentage of the explained variance
    component_loadings : pd.DataFrame
        Component (factor) loadings
    component_loadings_rotated : pd.DataFrame
        Component (factor) loadings after rotation
    communalities : pd.DataFrame
        Communalities
    communalities_and_loadings : pd.DataFrame
        A joint table of component (factor) loadings and communalities
    """
    
    def __init__(self, n_components='Kaiser', rotation=None, kappa=4):
        if str(n_components).lower() == 'kaiser':
            self.n_components_criterion = 'Kaiser'
        elif str(n_components).lower() == 'inflection':
            self.n_components_criterion = 'Inflection point (based on scree plot)'
        elif isinstance(n_components, int):
            self.n_components = n_components
            self.n_components_criterion = 'User based'
        else:
            raise ValueError(f"""Invalid number of components was passed.
            Possible values: exact number of components, 'Kaiser', or 'Inflection'.""")
            
        possible_rotations = ['varimax', 'natural collinearity', 'promax']
        
        if rotation is not None:
            if rotation.lower() not in possible_rotations:
                raise ValueError(f"Invalid type of rotation was passed. Possible values: {', '.join(possible_rotations)}.")            
            else:
                self.rotation = rotation.lower()
        else:
            self.rotation = None
                                
        self.kappa = kappa
        self._pc_max_list = None
                    
    def fit(self, 
            data, 
            variables=None, 
            scale=True, 
            show_results=True,
            n_decimals=3,
            print_decision=True):
        
        """
        Fit a model to the given data.
        
        Parameters
        ----------
        
        data : pd.DataFrame 
            Data to fit a model
            variables (None or list): variables from data to include in a model.
            If not specified, all variables will be used.
        variables : list
            Names of variables from data that should be used in a model.
            If not specified, all variables from data are used. Variables should have a numeric dtype.
        scale : bool 
            Whether data should be considered as scale variables. 
            If set to False, data will be transformed to ranks. 
        show_results :bool 
            Whether to show results of the analysis
        n_decimals : int 
            Number of digits to round results when showing them
        print_decision : bool
            Whether to print decision on what number of dimensions was exctracted

        Returns
        -------
        self
            The current instance of the PCA class
        """    
        
        if variables is not None:
            data = data[variables].dropna()
        else:
            data = data.dropna()
                                 
        if not scale:
            data = data.rank()
            
        self._data = data.copy()
        init_df = data.copy()
        
        self.variables = list(init_df.columns)
        self.max_n_components = len(self.variables)
        self._PCs_init_list = [f'PC{i+1}' for i in range(self.max_n_components)]
        self.correlation_matrix = pd.DataFrame(np.corrcoef(init_df.T),columns=init_df.columns,index=init_df.columns)
        self.correlation_matrix.index.name = 'Variable'
        self._eigenvectors, self._eigenvalues, self._eigenvectors_t = np.linalg.svd(self.correlation_matrix)
        self._eigenvalues, self._eigenvectors = PCA._sort_sign(self._eigenvalues, self._eigenvectors)
        
        #max_n_components solution
        #start = time.time()
        init_component_loadings = PCA._loadings(self._eigenvalues, 
                                          self._eigenvectors,
                                          self.variables,
                                          self._PCs_init_list)
        #print(f"._loadings(): {time.time()-start}")
        self.explained_variance = self.get_explained_variance(scree_plot=False)
        
        if self.n_components_criterion == 'Kaiser':
            
            self.n_components = len(self.explained_variance[self.explained_variance.iloc[:, 0].astype('float') >= 1])
#             if print_decision:
#                 print(f'The number of selected components by Kaiser criterion: {self.n_components}')
            
        elif self.n_components_criterion == 'Inflection point (based on scree plot)':
            self.n_components = PCA._inflection_point(self.explained_variance)
#             if print_decision:
#                 print(f'The number of selected components by an inflection point: {self.n_components}')
        self.explained_variance_total = self.explained_variance.loc[self.n_components, 'Cumulative %']
        
        #n_components solution
        self.component_loadings = init_component_loadings.iloc[:, :self.n_components]
        self.component_loadings_rotated = None
                             
        self._PCs_final_list = self._PCs_init_list[:self.n_components]
        
        if self.rotation is not None:
            if self.n_components > 1:
                self.component_loadings_rotated,self.structure_matrix = self._rotation_matrices(init_df,
                                                                                 self.n_components,
                                                                                 self.variables,
                                                                                 self._PCs_final_list,
                                                                                 self.component_loadings,
                                                                                 show_results)
    
                self._PCs_final_list = list(self.component_loadings_rotated.columns)
            else:
                print('Rotation could not be performed because number of dimensions is 1.')
                self.rotation = None

        
        if self.rotation in ['natural collinearity', 'promax']:
            necessary_loadings = self.structure_matrix.copy()
        elif self.rotation == 'varimax':
            necessary_loadings = self.component_loadings_rotated.copy()
        else:
            necessary_loadings = self.component_loadings.copy()
            
        #communalities through the component loadings
        self._necessary_loadings = necessary_loadings.copy()
        self.communalities = self.get_communalities(min_max=False)        
        
        self.communalities_and_loadings = pd.concat([self.communalities, necessary_loadings], axis=1)
        
        if self.rotation in ['natural collinearity', 'promax']:
            #start = time.time()     
            self._transformed_data_rotated = self.transform(init_df, add_to_data=False)
            #print(f".transform(): {time.time()-start}")        
            self.correlation_matrix_components = pd.DataFrame(np.corrcoef(self._transformed_data_rotated.T),
                                                          columns=self._transformed_data_rotated.columns,
                                                          index=self._transformed_data_rotated.columns)
            
            pc_corr = self.correlation_matrix_components.copy()
            _pc_max_list = []
            for pc in pc_corr:    
                _pc_max_list.append(abs(pc_corr[pc][(pc_corr[pc])!=1]).max())
            _pc_max_list.sort()
            self._pc_max_list = _pc_max_list.copy()
            
        if show_results:
            self.show_results()
                                 
        return self

    def show_results(self, print_decision=True, n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int 
            Number of digits to round results when showing them
        print_decision : bool
            Whether to print decision on what number of dimensions was exctracted
        """
        if self.rotation != 'natural collinearity':
            print('\nPCA SUMMARY')
            print('------------------\n')                                 
            if print_decision:
                if self.n_components_criterion=='Kaiser': 
                    print(f'The number of selected components by Kaiser criterion: {self.n_components}')
                elif self.n_components_criterion=='Inflection point (based on scree plot)':
                    print(f'The number of selected components by an inflection point: {self.n_components}')
                print('------------------\n')
        
        if self.rotation != 'natural collinearity':
            print('Explained variance')
            display(self.get_explained_variance(scree_plot=True).style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("methods .get_explained_variance() and .scree_plot()"))
            print(f'The model explains {round(self.explained_variance_total, n_decimals)}% of variance.')
            print('------------------\n')
        if self.rotation is None:
            print('Component loadings')
        elif self.rotation == 'varimax':
            print('Rotated component loadings')
        elif self.rotation in ['natural collinearity', 'promax']:
            print('Structure matrix')
        display(self.communalities_and_loadings.style\
                .format(None, na_rep="", precision=n_decimals)\
                .set_caption("attribute .communalities_and_loadings"))
        print(f'The minimum communality is {round(self.communalities["Communality"].min(), n_decimals)}.')
        if self.rotation in ['natural collinearity', 'promax']:
            print("Components' correlation")
            display(self.correlation_matrix_components.style\
                .format(None, na_rep="", precision=n_decimals)\
                .set_caption("attribute .correlation_matrix_components"))
        print('------------------\n')
        print('To get component scores, use [model].transform().')
    def transform(self, data, standardize=True, add_to_data=False):
        
        """
        Return component scores for every observation in the given dataset. 
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data to apply the model.
            If not specified, data that were used to fit the model will be used.
        standardize : bool 
            Whether to apply z-standartization to component scores
        add_to_data : bool 
            Whether to add variables of component scores to the given data

        Returns
        -------
        pd.DataFrame
            Requested values
        """
        
        df = data.copy()
        df = (df - df.mean()) / df.std()
        if self.rotation is not None and self.rotation == 'natural collinearity':
            return PCA._nat_collinearity(data, self.n_components, show=False, add_to_data=add_to_data, loadings_matrix=True)[3]
        else:                     
            loadings_df = self._necessary_loadings.copy()
                             
        df.columns = loadings_df.index
        try:
            weights = np.linalg.solve(self.correlation_matrix, loadings_df)
            #display(weights)
        except:
            weights = loadings_df.copy()
        component_scores_df = pd.DataFrame(np.dot(df, weights), 
                                           index = df.index,
                                           columns = loadings_df.columns)
        if standardize:
            component_scores_df = (component_scores_df - component_scores_df.mean()) / component_scores_df.std()
            
        if add_to_data:
            component_scores_df = pd.concat([df, component_scores_df], axis=1)

        return component_scores_df
        
    @staticmethod
    def _sort_sign(S, U):
        idx = S.argsort()[::-1]
        S = S[idx]
        U = U[:, idx]
        for i in range(U.shape[1]):
            if U[:, i].sum()<0:
                U[:, i] = -1 * U[:, i]
        return S, U
    
    @staticmethod
    def _loadings(S, U, vars_list, PCs_list):
        loadings = U * np.sqrt(S)
#         print('S',S,'np.sqrt(S)',np.sqrt(S))
        loadings_df = pd.DataFrame(loadings)
        loadings_df.columns = PCs_list
        loadings_df.index = vars_list
        loadings_df.index.name = 'Variable'
        return loadings_df
        
    def _rotation_matrices(self, data, n_components, vars_list, PCs_list, loadings_, show):
        if self.rotation == 'natural collinearity':
            matrices = PCA._nat_collinearity(data, n_components, show=show, add_to_data=False, loadings_matrix=True)
            component_loadings_rotated = matrices[1]
            structure_matrix = matrices[2]
            #print("rotation == 'natural collinearity'")

        else:
#             start = time.time()
            rotator = Rotator(vars_list, PCs_list, loadings_, kappa=self.kappa)
#             print(f"Rotator(): {time.time()-start}")        
            
            if self.rotation == 'varimax':
                loadings,rotation_mtx = rotator._varimax()            
#                 signs = np.sign(loadings.sum(0))
#                 signs[(signs==0)] = 1
#                 loadings = np.dot(loadings,np.diag(signs))
#                 component_loadings_rotated = pd.DataFrame(loadings,
#                                                  columns=loadings_.columns,
#                                                  index=loadings_.index)
                component_loadings_rotated = loadings
                structure_matrix = None
                #print("rotation == 'varimax'")
            
            else:
#                 start = time.time()
                loadings,rotation_mtx,phi = rotator._promax()
#                 print(f"_promax(): {time.time()-start}")
#                 signs = np.sign(loadings.sum(0))
#                 signs[(signs==0)] = 1
#                 loadings = np.dot(loadings,np.diag(signs))
#                 phi = np.dot(np.dot(np.diag(signs),phi),np.diag(signs))
                component_loadings_rotated = loadings
#                 start = time.time()
                structure_matrix = np.dot(loadings,phi)
#                 print(f".dot(): {time.time()-start}")
#                 start = time.time()
                structure_matrix = pd.DataFrame(structure_matrix,
                                                 columns=loadings.columns,
                                                 index=loadings.index)
#                 print(f".DataFrame(): {time.time()-start}")
                #print("rotation == 'promax'")

        return component_loadings_rotated,structure_matrix
    
    @staticmethod
    def _inflection_point(explained_variance, tol=0.05):
        percent_df = pd.DataFrame(columns=['Num','Percent'],
                                  index=range(1, len(explained_variance)+1))
        percent_df['Num'] = range(1, len(explained_variance)+1)
        percent_df['Percent'] = (explained_variance.iloc[:, 1]).astype('float')
        inflection_df = pd.DataFrame(columns=['delta'])
        for i in range(len(percent_df) - 1,0,-1): # running from the right point on the graph to the left
            results = smf.ols('Percent~Num', data=percent_df.loc[i:len(percent_df), :]).fit()
            a = results.rsquared
            results = smf.ols('Percent~Num', data=percent_df.loc[i - 1:len(percent_df), :]).fit()
            b = results.rsquared
            inflection_df.loc[i, 'delta'] = a - b

        inflection_point = inflection_df[inflection_df['delta'] == inflection_df['delta'].max()].index[0] - 1
        #print(f'The number of selected components by an inflection point = {inflection_point}')
        return inflection_point
#         delta = 0
#         i = len(percent_df) - 1
#         while delta < tol:
#             results = smf.ols('Percent ~ Num', data=percent_df.loc[i:len(percent_df), :]).fit()
#             a = results.rsquared
#             results = smf.ols('Percent ~ Num', data=percent_df.loc[i-1:len(percent_df), :]).fit()
#             b = results.rsquared
#             delta = a - b
#             i -= 1
        
#         return i
                             
    @staticmethod
    def _nat_collinearity(data, n_components, show=True, add_to_data=False, loadings_matrix=False):
        vrmx_model = PCA(n_components=n_components, rotation='varimax').fit(data, print_decision=False, show_results=False)                         
        dist_2 = vrmx_model.variables_components_distribution(show=False)[1]                         
        component_loadings = pd.DataFrame(index=data.columns)
        component_scores = pd.DataFrame(index=data.index)
        eigenvalue = 0
        i = 0
        for i_v in dist_2['Variables']:
            i += 1
            model = PCA().fit(data[i_v], print_decision=False, show_results=False)
            component_loadings[f'PC{i}_natcol'] = model.component_loadings['PC1']# component loadings -- МОГУТ ВЫВОДИТЬСЯ ПО СПЕЦ.ЗАПРОСУ
            eigenvalue += model.get_explained_variance(scree_plot=show if i < 5 else False)['Eigenvalue'][1]
            component_scores[f'PC{i}_natcol'] = model.transform(data[i_v])
        pc_corr = pd.DataFrame(np.corrcoef(component_scores.T),
                               columns=component_scores.columns,
                               index=component_scores.columns) # correlation of the components
        component_scores = pd.concat([data, component_scores],axis=1)
        structure_matrix = component_scores.corr()
        structure_matrix = structure_matrix.loc[data.columns, pc_corr.columns]

        if add_to_data==False:
            component_scores = component_scores.drop(data.columns, axis=1)

        if show:
            print(f'The total eigenvalue explained by Natural collinearity PCA (i.e., {len(dist_2)} independent PCA-models) = {round(eigenvalue,3)}',
            f'The total variance explained by Natural collinearity PCA = {round(100*eigenvalue/len(data.columns),3)}%',
            '', sep='\n')   

        if loadings_matrix:
            return pc_corr, component_loadings, structure_matrix, component_scores
        else:
            return pc_corr, structure_matrix, component_scores

    def get_communalities(self, min_max=True):
        
        """
        Return communalities for every initial variable. 
        
        Parameters
        ----------
        min_max : bool 
            Whether to print minimum and maximum of communalities

        Returns
        -------
        pd.DataFrame
            A table with communalities
        """
        loadings = self.component_loadings.copy()
        vars_list = self.variables.copy()
        PCs_list = self._PCs_final_list.copy()
        communalities = pd.DataFrame(index=loadings.index)
        communalities['Communality'] = (loadings**2).sum(axis=1)
        if min_max:
            print(f"The min communality: {round(communalities['Communality'].min(), 3)}, the max communality: {round(communalities['Communality'].max(), 3)}")
        return communalities
    
    def get_explained_variance(self, n_decimals=3, scree_plot=True):
        
        """
        Return summary table with information about variance accounted for. 
        
        Parameters
        ----------
        scree_plot : bool 
            Whether to display scree plot        
        annotate_bars : bool 
            Whether to annotate exact percentage of variance on each bar
            (if scree_plot set to True)
        annotate_current : bool 
            Whether to show percentage of variance corresponded 
            to the current solution
            (if scree_plot set to True)

        Returns
        -------
        pd.DataFrame
            A table with explained variance
        """          
                  
        S = self._eigenvalues.copy()
        PCs_list = self._PCs_init_list.copy()
        exp = 100 * S / np.sum(S)
        acc_sum = np.cumsum(exp)
        explained_variance = np.array([S, exp, acc_sum])
        explained_variance = pd.DataFrame(explained_variance.T, 
                                   columns=['Eigenvalue','Variance accounted for, %','Cumulative %'],
                                   index = range(1, len(PCs_list)+1))
        explained_variance.index.name = 'Component'
        explained_variance = round(explained_variance, n_decimals)

        if scree_plot:
            self.scree_plot()
        
        return explained_variance
    
    def scree_plot(self, annotate_bars=True, annotate_current=True):
                  
        """
        Vizualize distribution of the variance accounted for.  
        
        Parameters
        ----------
        annotate_bars : bool 
            Whether to annotate exact percentage of variance on each bar
        annotate_current : bool 
            Whether to show percentage of variance corresponded 
            to the current solution
        """ 
            
        explained_variance = self.explained_variance.copy()    
        plt.figure(figsize=(6, 6))
        
        pc_num = list(explained_variance.index)
        each_exp = explained_variance.iloc[:, 1].tolist()
        acc_sum = explained_variance.iloc[:, 2].tolist()
        plt.bar(pc_num, acc_sum, width=0.5, color='lightsalmon', alpha=0.2, label='Cumulative %')
        plt.plot(pc_num, each_exp, label='Variance accounted for, %')
        plt.plot(pc_num, each_exp, 'ro', label='_nolegend_')
        
        if annotate_bars:
            for x, y in zip(pc_num, acc_sum):
                plt.annotate(f'{round(y)}%', (x-0.25, y+0.75))
        
        if annotate_current:
            acc_sum_by_n_components = [explained_variance.iloc[self.n_components-1, 2]] * len(pc_num)
            plt.plot(pc_num, acc_sum_by_n_components, linestyle='--', label='Current solution', c='black')
                
        plt.xlabel('Principal components')
        plt.ylabel('Variance accounted for, %')
        plt.title('Variance accounted by components', fontsize=16)
        plt.legend(loc='upper left')
        plt.show()
        
    def variables_components_distribution(self,
                                          threshold=0.5,
                                          delta=0.1,
                                          show=True):
                  
        """
        Return several useful tables (as tuple) that may help to interpret solution:
        the distribution of variables through components,
        the distribution of components through variables,
        the competing components within variables.  
        
        Parameters
        ----------
        threshold : float 
            What value of an absolute component loading 
            to use when deciding about which variable belongs mostly to which component
            (default: 0.5)
        
        delta : float 
            What value of an absolute component loading 
            to consider as big enough to exclude component from competing
            (default: 0.1)
        
        show : bool 
            Whether to display all dataframes at once in a convinient way
            (default: True)

        Returns
        -------
        tuple
            A sequence of the requested dataframes
        """
        if self.n_components < 2:
            print('The distributions would not be shown because number of dimensions is 1.')
        else:
            loadings_df = self._necessary_loadings.copy()
            PCs_list = self._PCs_final_list.copy()

            PCs_through_vars = loadings_df.copy() # the components distribution through variables (loading >= 0.5 by module)
            for PC in PCs_list:
                PCs_through_vars[PC] = PCs_through_vars[PC].apply(lambda loading: '+' if abs(loading)>=threshold else '')

            vars_through_PCs = loadings_df.copy() # the variables distribution through components (loading >= 0.5 by module)
            vars_through_PCs_list = []
            for PC in PCs_list:
                vars_candidates_plus = list(vars_through_PCs[PC][vars_through_PCs[PC]>=threshold].index)
                vars_candidates_minus = list(vars_through_PCs[PC][vars_through_PCs[PC]<=-threshold].index)
                #if vars_candidates != []:
                vars_through_PCs_list.append([PC, vars_candidates_plus, vars_candidates_minus])
            vars_through_PCs = pd.DataFrame(vars_through_PCs_list)
            vars_through_PCs.columns = ['Component', 'Variables (positive side)', 'Variables (negative side)']
            vars_through_PCs.set_index('Component', inplace=True)

            #переименовать переменные для лучшей читаемости
            try:
                loadings_df_3 = loadings_df.copy() # the competing components within variables (loading >= 0.5 by module & difference < 0.1)
                for PC in PCs_list:
                    loadings_df_3[PC] = loadings_df_3[PC].apply(lambda loading: abs(loading) if abs(loading)>=threshold else None)
                loadings_df_3_1 = loadings_df_3.T
                v_PC_multi_list_3 = []
                for var in loadings_df_3_1.columns:
                    loadings_df_3_2 = loadings_df_3_1[var].copy()
                    loadings_df_3_2 = loadings_df_3_2.sort_values(ascending=False)
                    n_competing_components = loadings_df_3_2.count()
                    if n_competing_components > 1:
                        if loadings_df_3_2[0]-loadings_df_3_2[1] >= delta:
                            v_PC_multi_list_3.append([var, 
                                                      n_competing_components, 
                                                      (loadings_df_3_2[loadings_df_3_2 == loadings_df_3_2[0]].index)[0],
                                                      None])
                        else:
                        # БЕРУ ТОЛЬКО 2 КОНКУРИРУЮЩИЕ КОМПОНЕНТЫ (т.к. при вращении вряд ли потребуется больше)
                            try:
                                v_PC_multi_list_3.append([var, 
                                                          n_competing_components, 
                                                          (loadings_df_3_2[loadings_df_3_2 == loadings_df_3_2[0]].index)[0],
                                                          (loadings_df_3_2[loadings_df_3_2 == loadings_df_3_2[1]].index)[0]])
                            except:
                                v_PC_multi_list_3.append([var, 
                                                          0, 
                                                          None,
                                                          None])
                  
                    else:
                        try:
                            v_PC_multi_list_3.append([var, 
                                                  n_competing_components, 
                                                  (loadings_df_3_2[loadings_df_3_2 == loadings_df_3_2[0]].index)[0],
                                                  None])
                        except:
                            v_PC_multi_list_3.append([var, 
                                                          0, 
                                                          None,
                                                          None])
                v_PC_multi_df_3 = pd.DataFrame(v_PC_multi_list_3,
                                               columns=['Variable',
                                                        'N of competing components',
                                                        'Stronger component',
                                                        'Weaker component'])
                v_PC_multi_df_3.set_index('Variable', inplace=True)
            except:
                v_PC_multi_df_3 = None

            if show:
                print('The fist dataframe represents the distribution of variables across components.')
                display(PCs_through_vars)
                print('-------------')
                print('The second dataframe represents the distribution of components across variables.')      
                display(vars_through_PCs)
                print('-------------')
                print(f'''The third dataframe represents the competing components within variables 
                (which loadings >= {threshold} & difference between loadings <= {delta})''')
                display(v_PC_multi_df_3)

            return PCs_through_vars, vars_through_PCs, v_PC_multi_df_3


# previous version
# class PCA:
    
#     """
#     A class for implementing principal component analysis (PCA).
    
#     Parameters
#     ----------
#     n_components : int or str 
#         The exact number of dimensions
#         in solution or the criterion for its automatic selection.
#         Current possible values: 'kaiser' (defualt),
#         which corresponds to Kaiser's criterion
#     rotation : None or 'varimax' 
#         Rotation to perform on factor loadings.
#         Currently, only varimax rotation is available

#     Attributes
#     ----------
#     correlation_matrix : pd.DataFrame
#         A correlation matrix
#     explained_variance : pd.DataFrame
#         A table with eigenvalues and variance accounted for
#     explained_variance_total : float
#         Total percentage of the explained variance
#     component_loadings : pd.DataFrame
#         Component (factor) loadings
#     component_loadings_rotated : pd.DataFrame
#         Component (factor) loadings after rotation
#     communalities : pd.DataFrame
#         Communalities
#     communalities_and_loadings : pd.DataFrame
#         A joint table of component (factor) loadings and communalities
#     """
    
#     def __init__(self, n_components='Kaiser', rotation=None):
#         if str(n_components).lower() == 'kaiser':
#             self.n_components_criterion = 'Kaiser'
#         elif isinstance(n_components, int):
#             self.n_components = n_components
#             self.n_components_criterion = 'User based'
#         else:
#             raise ValueError(f"""Invalid number of components was passed.
#             Possible values: exact number of components or 'Kaiser'.""")
            
#         possible_rotations = ['varimax']
        
#         if rotation is not None:
#             if rotation.lower() not in possible_rotations:
#                 phrase = ', '.join(possible_rotations)
#                 raise ValueError(f"Invalid type of rotation was passed. Possible values: {phrase}.")

#             else:
#                 self.rotation = rotation.lower()
#         else:
#             self.rotation = None
                                
#         #self.kappa = kappa
#         #self._pc_max_list = None
                    
#     def fit(
#         self, 
#         data, 
#         variables=None,
#         scale=True,
#         show_results=True, 
#         n_decimals=3
#     ):
        
#         """
#         Fit a model to the given data.
        
#         Parameters
#         ----------
        
#         data : pd.DataFrame 
#             Data to fit a model
#             variables (None or list): variables from data to include in a model.
#             If not specified, all variables will be used.
#         variables : list
#             Names of variables from data that should be used in a model.
#             If not specified, all variables from data are used. Variables should have a numeric dtype.
#         scale : bool 
#             Whether data should be considered as scale variables. 
#             If set to False, data will be transformed to ranks. 
#         show_results :bool 
#             Whether to show results of the analysis
#         n_decimals : int 
#             Number of digits to round results when showing them

#         Returns
#         -------
#         self
#             The current instance of the PCA class
#         """    
        
#         if variables is not None:
#             data = data[variables].dropna()
#         else:
#             data = data.dropna()
                                 
#         if not scale:
#             data = data.rank()
                                 
#         self._data = data.copy()
#         self.variables = list(data.columns)
#         self.max_n_components = len(self.variables)
#         self._PCs_init_list = [f'PC{i+1}' for i in range(self.max_n_components)]
#         self.correlation_matrix = pd.DataFrame(
#             np.corrcoef(data.T),
#             columns=data.columns,
#             index=data.columns
#         )
#         self.correlation_matrix.index.name = 'Variable'
                                 
#         self._eigenvectors, self._eigenvalues, self._eigenvectors_t = np.linalg.svd(self.correlation_matrix)
#         self._eigenvalues, self._eigenvectors = PCA._sort_sign(self._eigenvalues, self._eigenvectors)
        
#         #max_n_components solution
#         init_component_loadings = PCA._loadings(self._eigenvalues, 
#                                           self._eigenvectors,
#                                           self.variables,
#                                           self._PCs_init_list)
                                 
#         self.explained_variance = self.get_explained_variance(scree_plot=False)
        
#         if self.n_components_criterion == 'Kaiser':
            
#             self.n_components = len(self.explained_variance[self.explained_variance.iloc[:, 0].astype('float') >= 1])
                                 
#         self.explained_variance_total = self.explained_variance.loc[self.n_components, 'Cumulative %']
        
#         #n_components solution
#         self.component_loadings = init_component_loadings.iloc[:, :self.n_components]
#         self.component_loadings_rotated = None
                             
#         self._PCs_final_list = self._PCs_init_list[:self.n_components]
        
#         if self.rotation is not None:
#             if self.n_components > 1:
#                 self.component_loadings_rotated, self.structure_matrix = self._rotation_matrices()
    
#                 self._PCs_final_list = list(self.component_loadings_rotated.columns)
#             else:
#                 print('Rotation could not be performed because number of dimensions is 1.')
#                 self.rotation = None

#         if self.rotation == 'varimax':
#             necessary_loadings = self.component_loadings_rotated.copy()
#         else:
#             necessary_loadings = self.component_loadings.copy()
            
#         #communalities through the component loadings
#         self._necessary_loadings = necessary_loadings.copy()
#         self.communalities = self.get_communalities(min_max=False)        
        
#         self.communalities_and_loadings = pd.concat([necessary_loadings, self.communalities], axis=1)
            
#         if show_results:
#             self.show_results(n_decimals=n_decimals)
                                 
#         return self

#     def show_results(self, n_decimals=3):
#         """
#         Show results of the analysis in a readable form.
        
#         Parameters
#         ----------
#         n_decimals : int 
#             Number of digits to round results when showing them
#         """
#         print('\nPCA SUMMARY')
#         print('------------------\n')                                 
#         if self.n_components_criterion=='Kaiser':
#             print(f'The number of selected components by Kaiser criterion: {self.n_components}')
#             print('------------------\n')
#         print('Explained variance')
#         display(self.get_explained_variance(scree_plot=True).style\
#                     .format(None, na_rep="", precision=n_decimals)\
#                     .set_caption("methods .get_explained_variance() and .scree_plot()"))
#         print(f'The model explains {round(self.explained_variance_total, 3)}% of variance.')
#         print('------------------\n')
#         if self.rotation is None:
#             print('Component loadings')
#         elif self.rotation == 'varimax':
#             print('Rotated component loadings')
#         display(self.communalities_and_loadings.style\
#                 .format(None, na_rep="", precision=n_decimals)\
#                 .set_caption("attribute .communalities_and_loadings"))
#         print(f'The minimum communality is {round(self.communalities["Communality"].min(), 3)}.')
#         print('------------------\n')
#         print('To get component scores, use [model].transform().')

#     def transform(self, data=None, standardize=True, add_to_data=False):
        
#         """
#         Return component scores for every observation in the given dataset. 
        
#         Parameters
#         ----------
#         data : pd.DataFrame 
#             Data to apply the model.
#             If not specified, data that were used to fit the model will be used.
#         standardize : bool 
#             Whether to apply z-standartization to component scores
#         add_to_data : bool 
#             Whether to add variables of component scores to the given data

#         Returns
#         -------
#         pd.DataFrame
#             Requested values
#         """
#         if data is None:
#             data = self._data.copy()
#             df = self._data.copy()
#         else:
#             df = data[self.variables].dropna().copy()
              
#         df = (df - df.mean()) / df.std()
#         loadings_df = self._necessary_loadings.copy()                     
#         df.columns = loadings_df.index
                  
#         try:
#             weights = np.linalg.solve(self.correlation_matrix, loadings_df)
#         except:
#             weights = loadings_df.copy()
#         component_scores_df = pd.DataFrame(np.dot(df, weights), 
#                                            index = df.index,
#                                            columns = loadings_df.columns)
#         if standardize:
#             component_scores_df = (component_scores_df - component_scores_df.mean()) / component_scores_df.std()
            
#         if add_to_data:
#             component_scores_df = pd.concat([data, component_scores_df], axis=1)

#         return component_scores_df
        
#     @staticmethod
#     def _sort_sign(S, U):
#         idx = S.argsort()[::-1]
#         S = S[idx]
#         U = U[:, idx]
#         for i in range(U.shape[1]):
#             if U[:, i].sum()<0:
#                 U[:, i] = -1 * U[:, i]
#         return S, U
    
#     @staticmethod
#     def _loadings(S, U, vars_list, PCs_list):
#         loadings = U * np.sqrt(S)
#         loadings_df = pd.DataFrame(loadings)
#         loadings_df.columns = PCs_list
#         loadings_df.index = vars_list
#         loadings_df.index.name = 'Variable'
#         return loadings_df
        
#     def _rotation_matrices(self):
#         if self.rotation == 'varimax':
#             rot = Rotator(method='varimax', normalize=True)
#             loadings = rot.fit_transform(self.component_loadings)
#             loadings = pd.DataFrame(loadings,
#                                    columns=[f'PC{i+1}_vrmx' for i in range(self.n_components)],
#                                    index=self.variables)
#             structure_matrix = None

#         return loadings, structure_matrix


#     def get_communalities(self, min_max=True):
        
#         """
#         Return communalities for every initial variable. 
        
#         Parameters
#         ----------
#         min_max : bool 
#             Whether to print minimum and maximum of communalities

#         Returns
#         -------
#         pd.DataFrame
#             A table with communalities
#         """
#         loadings = self.component_loadings.copy()
#         communalities = pd.DataFrame(index=loadings.index)
#         communalities['Communality'] = (loadings**2).sum(axis=1)
         
#         if min_max:
#             min_ = round(communalities['Communality'].min(), 3)
#             max_ = round(communalities['Communality'].max(), 3)
#             print(f"The min communality: {min_}, the max communality: {max_}")
        
#         return communalities
    
#     def get_explained_variance(self, scree_plot=True, **kwargs):
        
#         """
#         Return summary table with information about variance accounted for. 
        
#         Parameters
#         ----------
#         scree_plot : bool 
#             Whether to display scree plot        
#         annotate_bars : bool 
#             Whether to annotate exact percentage of variance on each bar
#             (if scree_plot set to True)
#         annotate_current : bool 
#             Whether to show percentage of variance corresponded 
#             to the current solution
#             (if scree_plot set to True)

#         Returns
#         -------
#         pd.DataFrame
#             A table with explained variance
#         """          
                  
#         S = self._eigenvalues
#         PCs_list = self._PCs_init_list
#         exp = 100 * S / np.sum(S)
#         acc_sum = np.cumsum(exp)
#         explained_variance = np.array([S, exp, acc_sum])
#         explained_variance = pd.DataFrame(explained_variance.T, 
#                                    columns=['Eigenvalue','Variance accounted for, %','Cumulative %'],
#                                    index = range(1, len(PCs_list)+1))
#         explained_variance.index.name = 'Component'

#         if scree_plot:
#             self.scree_plot(**kwargs)
        
#         return explained_variance
    
#     def scree_plot(self, annotate_bars=True, annotate_current=True):
                  
#         """
#         Vizualize distribution of the variance accounted for.  
        
#         Parameters
#         ----------
#         annotate_bars : bool 
#             Whether to annotate exact percentage of variance on each bar
#         annotate_current : bool 
#             Whether to show percentage of variance corresponded 
#             to the current solution
#         """ 
            
#         explained_variance = self.explained_variance.copy()    
#         plt.figure(figsize=(6, 6))
        
#         pc_num = list(explained_variance.index)
#         each_exp = explained_variance.iloc[:, 1].tolist()
#         acc_sum = explained_variance.iloc[:, 2].tolist()
#         plt.bar(pc_num, acc_sum, width=0.5, color='lightsalmon', alpha=0.2, label='Cumulative %')
#         plt.plot(pc_num, each_exp, label='Variance accounted for, %')
#         plt.plot(pc_num, each_exp, 'ro', label='_nolegend_')
        
#         if annotate_bars:
#             for x, y in zip(pc_num, acc_sum):
#                 plt.annotate(f'{round(y)}%', (x-0.25, y+0.75))
        
#         if annotate_current:
#             acc_sum_by_n_components = [explained_variance.iloc[self.n_components-1, 2]] * len(pc_num)
#             plt.plot(pc_num, acc_sum_by_n_components, linestyle='--', label='Current solution', c='black')
                
#         plt.xlabel('Principal components')
#         plt.ylabel('Variance accounted for, %')
#         plt.title('Variance accounted by components', fontsize=16)
#         plt.legend(loc='upper left')
#         plt.show()



# Авторский модуль для операций с эмитентами торгуемых на МосБирже облигаций

# 0. Активировать требуемые для работы скрипта модули и пакеты 
# sys & subprocess -- эти пакеты должны быть предустанавлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
while True:
    try:
        from randan.tools import coLabAdaptor, textPreprocessor # авторские модули для
        # (а) адаптации текущего скрипта к файловой системе CoLab и
        # (б) предобработки нестандартизированнрого текста
        import os, pandas, re
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[0]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

coLabFolder = coLabAdaptor.coLabAdaptor()

# Авторские функции..
    # .. компановки информации об эмитентах торгуемых на МосБирже облигаций в датафрейм (словарь)
def issuersComposer(bondS):
    issuerS = bondS[['Эмитент']].sort_values('Эмитент').drop_duplicates().reset_index(drop=True) # заготовка для датафрейма с актуальными эмитентами
    for row_issuerS in issuerS.index:
    # for row_issuerS in issuerS.index[:1]: # для отладки
        # print('row_issuerS:', row_issuerS) # для отладки
        rowS_bondS = bondS[bondS['Эмитент'] == issuerS['Эмитент'][row_issuerS]].index # обрабатываемые строчки bondS
        secNameS = bondS['SECNAME'][rowS_bondS].tolist()
        # print('secNameS:', secNameS) # для отладки
        issuerS.loc[row_issuerS, 'Count'] = len(secNameS)
        issuerS.loc[row_issuerS, 'SecNameS'] = ''
        issuerS.at[row_issuerS, 'SecNameS'] = secNameS

        if 'Bond D Rating' in bondS.columns:
            ratingS = bondS['Bond D Rating'][rowS_bondS].dropna().tolist()
            # print('ratingS:', ratingS) # для отладки
            ratingS_mean = bondS['Bond D Rating'][rowS_bondS].dropna().mean()
            # print('ratingS_mean:', ratingS_mean) # для отладки
            issuerS.loc[row_issuerS, 'RatingS'] = ''
            issuerS.at[row_issuerS, 'RatingS'] = ratingS[0] if len(ratingS) == 1 else ratingS
            issuerS.loc[row_issuerS, 'Issuer D Rating'] = ratingS_mean

    # display('issuerS:', issuerS) # для отладки
    return issuerS

    # .. извлечения из SECNAME торгуемых на МосБирже облигаций названий их эмитентов
def issuerExtractor(dfIn):
    df = dfIn.copy()
    df['Эмитент'] = df['SECNAME'].str.replace('_', ' ').str.replace('-', ' ')
    df['Эмитент'] = df['Эмитент'].str.replace('ЯНАО', 'ЯНАвОк')
    df.loc[df['Эмитент'].str.contains('ИОС'), 'Эмитент'] = 'СберИОС'
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: re.sub(r' [БЗ][OОPР]П?.+', '', cellContent))
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: re.sub(r'Б\d+', '', cellContent))
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: re.sub(r' 0\d+.*', '', cellContent))
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: re.sub(r' \d+ обл\.?', '', cellContent))
    df['Эмитент'] = df['Эмитент'].apply(textPreprocessor.simbolsCleaner)
    df['Эмитент'] = df['Эмитент'].str.replace('ПАО ', '').str.replace(' ПАО', '')\
        .str.replace('АО ', '').str.replace(' АО', '')\
        .str.replace('ООО ', '').str.replace(' ООО', '')\
        .str.replace('"', '')
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: cellContent if cellContent[0] != ' ' else cellContent[1:])
    df['Эмитент'] = df['Эмитент'].apply(lambda cellContent: cellContent if cellContent[-1] != ' ' else cellContent[:-1])
    df['Эмитент'] = df['Эмитент'].str.split(' Б ').str[0]
    df['Эмитент'] = df['Эмитент'].str.lower()
    df.loc[df['Эмитент'].str.contains('офз'), 'Эмитент'] = 'минфин рф'
    df.loc[(df['Эмитент'].str.contains('воз')) & (df['Эмитент'].str.contains('рф')), 'Эмитент'] = 'минфин рф'
    return df



from randan.tools import varPreprocessor
df = df.drop([904, 2173, 2392, 2822, 3252, 3558, 3865, 3925])
# Проверка всех столбцов на наличие в их ячейках JSON-формата
from randan.tools import varPreprocessor
columnsToJSON = varPreprocessor.jsonChecker(df)

if len(columnsToJSON) > 0:
    print('Обнаружены столбцы, содержащие внутри своих ячеек JSON-объекты; Excel не поддерживает JSON-формат;'
          , 'чтобы формат JSON не потерялся, сохраняю эти столбцы в файл формата НЕ XLSX, а JSON. Остальные же столбцы сохраняю в файл формата XLSX')

    columnsToJSON.append('id')
    if 'from_id' in df.columns: columnsToJSON.append('from_id')
    if 'owner_id' in df.columns: columnsToJSON.append('owner_id')

    df[columnsToJSON].to_json(r"YT_раздельный сбор\20250311_0230_YT_раздельный сбор_videos_JSON_varS.json")

    columnsToJSON.remove('id')
    if 'from_id' in columnsToJSON: columnsToJSON.remove('from_id')
    if 'owner_id' in columnsToJSON: columnsToJSON.remove('owner_id')

    df.drop(columnsToJSON, axis=1).to_excel(r"YT_раздельный сбор\20250311_0230_YT_раздельный сбор_videos_Other_varS.xlsx")



# Авторский модуль для выгрузки характеристик торгуемых на МосБирже облигаций

# 0. Активировать требуемые для работы скрипта модули и пакеты 
# sys & subprocess -- эти пакеты должны быть предустанавлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
while True:
    try:
        from bs4 import BeautifulSoup
        from tqdm import tqdm
        from randan.tools import coLabAdaptor # авторский модуль для адаптации текущего скрипта к файловой системе CoLab
        import os, pandas, re, requests, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[0]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

coLabFolder = coLabAdaptor.coLabAdaptor()

# 1. Авторские функции..
    # .. выгрузки имён полей БД МосБиржи
def getColumnNameS(text):
    columnS = BeautifulSoup(text, features='xml').find_all('column')
    # print('columnS:', columnS) # для отладки
    columnNameS = []
    for column in columnS:
        columnNameS.append(column.get('name'))
    return columnNameS

    # .. выгрузки данных из БД МосБиржи
def pseudojson2df(headerS, index, url):
    df = pandas.DataFrame()
    text = re.findall(r'<data.+?/data>', requests.get(url, headers=headerS).text, re.DOTALL)[index]
    # print('text:', text) # для отладки
    columnNameS = getColumnNameS(text)
    rowS = BeautifulSoup(text, features='xml').find_all('row')
    # print('rowS:', rowS) # для отладки
    i = 0
    for row in tqdm(rowS):
        i += 1
        for column in columnNameS:
            df.loc[i, column] = row.get(column)
    # display('df:', df)
    return df

# 2. Авторская функция исполнения скрипта
def getMoExData(
                market='bonds',
                path=coLabFolder,
                returnDfs=False
                ):
    """
    Функция умеет выгружать характеристики торгуемых на МосБирже облигаций, причём не дефолтные (далее -- Д) и не повышенного инвестиционного риска (далее -- ПИР). Дополнительно выгружается словарь полей БД МосБиржи. Также она умеет выгружать фьючерсы

    Parameters
    ----------
       market : str -- если интересуют облигации, подходит значение по умолчанию 'bonds' , если фьючерсы, впишите 'forts'

         path : str -- путь к директории, включая её имя, в которой будут искаться файлы и куда будут сохраняться; по умолчанию не в CoLab поиск и сохранение происходят в директории, в которой вызывается текущая функция, а в CoLab в директории Colab Notebooks

    returnDfs : bool -- в случае True функция возвращает итоговые датафреймы boardS, columnsDescriptionS и securitieS строго в такой последовательности
    """
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    if path == None: path = ''
    else: path += slash

# Формирование файла с режимами торгов
# 1.0 Если нет файла с режимами торгов    
    print('Создаю файл с режимами торгов')
    if market == 'bonds': url = f'https://iss.moex.com/iss/engines/stock/markets/{market}'
    if market == 'forts': url = f'https://iss.moex.com/iss/engines/futures/markets/{market}'
    headerS = {'Cookie': 'yashr=7199406881722422993; yabs-sid=1989516261722422994; gdpr=0; _ym_uid=172242390960576307; _ym_d=1722423909; yandex_login=aleksei.rotmistrov; yandexuid=1251707911713359739; yuidss=1251707911713359739; ymex=2038826180.yrts.1723466180; skid=98896631723495108; yabs-dsp=mts_banner.bjhrRmhRcmRTYWFqa2szTTdWRHB2UQ==; my=YwA=; amcuid=9218374081731017878; yandex_gid=213; is_gdpr=0; is_gdpr_b=COXEFxCCoAIoAg==; i=uleIuerZ29JaTX59z5G/+HKk9fEmnUoKXjW/KGLZiTQaKYElKHEzfPDCABcpPVUVz6h+GEzjHO3ElrWjkRmIAGlp+lY=; Session_id=3:1733758173.5.0.1730723942669:tOnmRptGBpQAvmusaCECKg:471c.1.2:1|454550616.-1.2.3:1730723942|3:10299425.956559.MUCs35YHhfnyWe6-GuWX5wjaRxs; sessar=1.1196.CiDL7YFrdyEcpmiO9V7a1ylcpw6ej8qZiLU8_AgTxsNW_w.AGdGhxY1_HuHtpuOQLQHoSH6QAM9RilP9yNVtHdZlXc; sessionid2=3:1733758173.5.0.1730723942669:tOnmRptGBpQAvmusaCECKg:471c.1.2:1|454550616.-1.2.3:1730723942|3:10299425.956559.fakesign0000000000000000000; _ym_isad=2; yabs-vdrf=A0; yp=1736202372.atds.1#1735233224.hdrc.0#2049118172.pcs.0#1765295017.swntab.0#1746491882.szm.1_875%3A1280x720%3A1676x760#2046083942.udn.cDphbGVrc2VpLnJvdG1pc3Ryb3Y%3D#1738499314.vhstfltr_onb.1%3A1730723314256#1734471330.ygu.1#1734622172.dlp.2; ys=udn.cDphbGVrc2VpLnJvdG1pc3Ryb3Y%3D#wprid.1733759017472236-15201027122329898332-balancer-l7leveler-kubr-yp-sas-248-BAL#c_chck.3523091273; bh=EkAiTWljcm9zb2Z0IEVkZ2UiO3Y9IjEzMSIsIkNocm9taXVtIjt2PSIxMzEiLCJOb3RfQSBCcmFuZCI7dj0iMjQiGgUieDg2IiIPIjEzMS4wLjI5MDMuODYiKgI/MDICIiI6CSJXaW5kb3dzIkIIIjEwLjAuMCJKBCI2NCJSXCJNaWNyb3NvZnQgRWRnZSI7dj0iMTMxLjAuMjkwMy44NiIsIkNocm9taXVtIjt2PSIxMzEuMC42Nzc4LjEwOSIsIk5vdF9BIEJyYW5kIjt2PSIyNC4wLjAuMCIiWgI/MA=='
        , 'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Edg/131.0.0.0'}
    boardS = pseudojson2df(headerS, 0, url)        
    # display('boardS:', boardS) # для отладки

# 1.1 Если облигации: нужны именно торгуемые и не Д , и не ПИР
    if market == 'bonds':
        boardS['is_traded'] = boardS['is_traded'].astype(int)
        boardS = boardS[
            (boardS['is_traded'] == 1) & (boardS['title'].str.contains('облигации ', case=False))\
            & (boardS['title'].str.contains('ПИР ') != True) & (boardS['title'].str.contains('Д ') != True)
                        ]
    # display('boardS:', boardS) # для отладки

# Формирование файла с доступными облигациями

# 2.1 Формирование словаря полей БД МосБиржи
    print('Создаю файл со словарём полей БД МосБиржи')
    columnsDescriptionS = pandas.DataFrame()
    if market == 'bonds': indeceS = [2, 8]
    if market == 'forts': indeceS = [2, 3]
    for index in indeceS:
        columnsDescriptionS_additional = pseudojson2df(headerS, index, url)
        columnsDescriptionS_additional.loc[:, 'data id'] = index
        columnsDescriptionS = pandas.concat([columnsDescriptionS, columnsDescriptionS_additional], ignore_index=True)
    columnsDescriptionS = columnsDescriptionS.drop_duplicates(['id', 'name'], ignore_index=True)
    # display('columnsDescriptionS:', columnsDescriptionS) # для отладки
    columnsDescriptionS.to_excel(market + 'ColumnsDescriptionS.xlsx', index=False)

# 2.2 Формирование файла с доступными облигациями в интересующих режимах торгов
    decision = ''
    if os.path.exists(path + market + 'SecuritieS.xlsx'):
        print(
'''--- Если НЕ хотите обновить файл с доступными инструментами в интересующих режимах торгов, просто нажмите Enter
--- Если хотите, то нажмите пробел и затем Enter'''
              )
        decision = input()

    if (os.path.exists(path + market + 'SecuritieS.xlsx') != True) | len(decision) != 0:
        print('Создаю файл с доступными инструментами в интересующих режимах торгов')
        securitieS = pandas.DataFrame()
        marketdata_yieldS = pandas.DataFrame()
        marketdata = pandas.DataFrame()
        # display("boardS['boardid']:", boardS['boardid']) # для отладки        
        for board in boardS['boardid']:
        # for board in ['TQOB']: # для отладки   
            print('board:', board)
            securitieS_additional = pseudojson2df(headerS, 0, url + f'/boards/{board}/securities')
            securitieS = pandas.concat([securitieS, securitieS_additional], ignore_index=True)

            if market == 'bonds':
                marketdata_yieldS_additional = pseudojson2df(headerS, -1, url + f'/boards/{board}/securities')
                marketdata_yieldS = pandas.concat([marketdata_yieldS, marketdata_yieldS_additional], ignore_index=True)
                # print('marketdata_yieldS.columns:', marketdata_yieldS.columns) # для отладки   

            if market == 'forts':
                marketdata_additional = pseudojson2df(headerS, 1, url + f'/boards/{board}/securities')
                # display('marketdata_additional:', marketdata_additional) # для отладки
                marketdata = pandas.concat([marketdata, marketdata_additional], ignore_index=True)
                # print('marketdata.columns:', marketdata.columns) # для отладки   

        if os.path.exists(path + market + 'ColumnsDescriptionsSelected.xlsx'):
            columnsDescriptionS = pandas.read_excel(path + market + 'ColumnsDescriptionsSelected.xlsx')
        else:
            columnsDescriptionS = pandas.read_excel(path + market + 'ColumnsDescriptionS.xlsx')
            # display('columnsDescriptionS:', columnsDescriptionS) # для отладки
            columnsDescriptionS = columnsDescriptionS[columnsDescriptionS['name'] !='BOARDID']

        columnsDescriptionS = columnsDescriptionS[columnsDescriptionS['name'].notna()]
        columnsDescriptionS = columnsDescriptionS['name'].drop_duplicates().tolist()
        if market == 'bonds': columnsDescriptionS.append('URL MoEx')

        if market == 'bonds':
            securitieS = securitieS.merge(marketdata_yieldS, on='SECID', suffixes=("", "_drop"), how="left")
            securitieS = securitieS[[column for column in securitieS.columns if not column.endswith("_drop")]]
            # print('securitieS.columns:', securitieS.columns) # для отладки

        if market == 'forts':
            securitieS = securitieS.merge(marketdata, on='SECID', suffixes=("", "_drop"), how="left")
            securitieS = securitieS[[column for column in securitieS.columns if not column.endswith("_drop")]]
            # print('securitieS.columns:', securitieS.columns) # для отладки

        securitieS = securitieS.groupby('SECID', as_index=False).first()
        # print('securitieS.columns:', securitieS.columns) # для отладки
        if market == 'bonds': securitieS['URL MoEx'] = 'https://www.moex.com/ru/issue.aspx?code=' + securitieS['ISIN']
        # print('securitieS.columns:', securitieS.columns) # для отладки
        securitieS = securitieS[columnsDescriptionS]
        securitieS.to_excel(path + market + 'SecuritieS.xlsx', index=False)
        # display(securitieS) # для отладки
    else:
        print('Файл с доступными инструментами в интересующих режимах торгов НЕ обновлялся')
        securitieS = pandas.read_excel(path + market + 'SecuritieS.xlsx')

    if returnDfs: return boardS, columnsDescriptionS, securitieS
    warnings.filterwarnings("ignore")
    print('Скрипт исполнен. Сейчас появится надпись: "An exception has occurred, use %tb to see the full traceback.\nSystemExit" -- так и должно быть')
    input()
    sys.exit()
# https://iss.moex.com/iss/reference/
# https://iss.moex.com/iss/engines/stock/markets/qnv
# https://iss.moex.com/iss/engines/stock/markets/bonds/boards/tqcb/securities
